"""A General Operation"""

from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._component import _Component
from ...modeling.parameters.conversion import Conversion

if TYPE_CHECKING:
    from ...modeling.constraints.calculate import Calculate
    from ..commodity.resource import Resource
    from ..spatial.linkage import Linkage
    from ..spatial.location import Location
    from ..temporal.lag import Lag
    from ..temporal.periods import Periods


@dataclass
class _Operation(_Component):
    """A General Operation

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: An optional citation or description for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conv: Operational conversion associated with the operation. Defaults to None.
    :vartype conv: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar fab: Material conversion associated with the operation. Defaults to None.
    :vartype fab: Conversion, optional
    :ivar _fab_balanced: True if the material conversion has been balanced. Defaults to False.
    :vartype _fab_balanced: bool
    """

    def __post_init__(self):
        _Component.__post_init__(self)

        self._conv = False

        # Operational conversion
        self.conv: Conversion = None

        # Material conversion
        self._fab: Conversion = None

        # to check if fab is balanced
        self._fab_balanced: bool = False

    @property
    @abstractmethod
    def spaces(self) -> list[Location | Linkage]:
        """Locations at which the process is balanced"""

    @property
    def fab(self) -> Conversion:
        """Material conversion"""

        if self._fab is None:
            # will be made the first time it is called
            self._fab = Conversion(operation=self, bind=self.capacity)
        return self._fab

    @property
    def capex(self) -> Calculate:
        """Capital Expenditure"""
        return self.capacity[self.model.default_currency().spend]

    @property
    def opex(self) -> Calculate:
        """Operational Expenditure"""
        return self.operate[self.model.default_currency().spend]

    @property
    def base(self) -> Resource:
        """Base resource"""
        return self.conv.base

    @property
    def conversion(self) -> dict[Resource, int | float]:
        """Conversion of commodities"""
        return self.conv.conversion

    @property
    def fabrication(
        self,
    ) -> (
        dict[Resource, int | float | list[int | float]]
        | dict[int | str, dict[Resource, int | float | list[int | float]]]
    ):
        """Material conversion of commodities"""
        if self.fab.pwl:
            return {
                mode: self.fab.conversion[_mode]
                for mode, _mode in zip(self.fab.modes, list(self.fab.conversion))
            }

        return self.fab.conversion

    @property
    def lag(self) -> Lag:
        """Lag of the process"""
        return self.conv.lag

    def writecons_fabrication(
        self,
        space_times: list[tuple[Location | Linkage, Periods]],
        # fabrication: dict[Resource, int | float | list[int | float]],
    ):
        """write fabrication constraints for the operation"""
        if not self._fab_balanced:
            self.fab.balancer()
            self._fab_balanced = True

        if self.fab.pwl:
            # n_modes = len(self.fabrication)
            # modes_name = f'bin{len(self.model.modes)}'

            # setattr(self.model, modes_name, Modes(n_modes=n_modes, bind=self.capacity))

            # modes = self.model.modes[-1]

            # this will create modes if not already created
            modes = self.fab.modes

        for space_time in space_times:
            space = space_time[0]
            time = space_time[1]

            if space in self.spaces:
                continue

        def time_checker():
            # write use on the densest temporal index
            return max(self.model.dispositions[self.model.capacity][self][space])

        if self.fab.pwl:

            _modes = list(self.fabrication)
            for n, mode in enumerate(modes):

                for res, par in self.fabrication[_modes[n]].items():

                    if isinstance(par, (int | float)):
                        if par == 0:
                            continue

                    if isinstance(par, list):
                        if par[0] == 0:
                            continue

                    _ = self.capacity(space, time_checker(), mode)[res.use(mode)] == par

        else:

            for res, par in self.fabrication.items():

                if isinstance(par, (int | float)):
                    if par == 0:
                        continue

                if isinstance(par, list):
                    if par[0] == 0:
                        continue

                if res in self.model.grb:
                    time = time_checker()

                    if self.model.grb[res][space][time]:
                        _insitu = False
                    else:
                        _insitu = True

                else:
                    _insitu = True

                if _insitu:
                    res.insitu = True
                    _ = res.use(self.capacity, space, time) == True

                _ = self.capacity(space, time)[res.use] == par

    def locate(self, *spaces: Location | Linkage):
        """Locate the process"""

        # get location, time tuples where operation is defined
        space_times: list[tuple[Location | Linkage, Periods]] = []
        for space in spaces:

            if self not in self.model.capacity.bound_spaces:
                self.model.capacity.bound_spaces[self] = {"ub": [], "lb": []}

            if space not in self.model.capacity.bound_spaces[self]["ub"]:
                # check if operational capacity has been bound
                print(
                    f"--- Assuming  {self} capacity is unbounded in ({space}, {self.horizon})",
                )
                # this is not a check, this generates a constraint
                _ = self.capacity(space, self.horizon) == True

            if self not in self.model.operate.bound_spaces:
                self.model.operate.bound_spaces[self] = {"ub": [], "lb": []}

            if space not in self.model.operate.bound_spaces[self]["ub"]:
                # check if operate has been bound
                # if not just write opr_{pro, space, horizon} <= capacity_{pro, space, horizon}
                print(
                    f"--- Assuming operation of {self} is bound by capacity in ({space}, {self.horizon})",
                )
                if (
                    self in self.model.operate.dispositions
                    and space in self.model.operate.dispositions[self]
                ):

                    _ = (
                        self.operate(
                            space,
                            min(self.model.operate.dispositions[self][space]),
                        )
                        <= 1
                    )
                else:
                    _ = self.operate(space, self.horizon) <= 1

            # check if the process is being operated at the location
            for d in self.model.operate.domains:
                if d.operation == self and d.space == space:
                    space_time = (space, d.time)
                    if space_time not in space_times:
                        space_times.append(space_time)

        self.writecons_conversion(space_times)

        if self.fabrication:
            self.writecons_fabrication(space_times)



"""Bind constraint"""

from __future__ import annotations

import time as keep_time
from typing import TYPE_CHECKING

from ...components.temporal.modes import Modes
from ...utils.math import normalize

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ..._core._component import _Component
    from ..._core._x import _X
    from ..variables.sample import Sample


class Bind:
    """Bind constraint

    :param sample: The sample variable to bind
    :type sample: Sample
    :param bound: The bound value
    :type bound: float | list[float] | dict[float, float] | tuple[float, float] | list[tuple[float, float]]
    :param leq: If True, the sample is constrained to be less than or equal to the bound
    :type leq: bool
    :param geq: If True, the sample is constrained to be greater than or equal to the bound
    :type geq: bool
    :param eq: If True, the sample is constrained to be equal to the bound
    :type eq: bool
    :param forall: If provided, the constraint is applied for all elements in this list
    :type forall: list[_X | _Component] | None
    """

    def __init__(
        self,
        sample: Sample,
        parameter: (
            float
            | list[float]
            | dict[float, float]
            | tuple[float, float]
            | list[tuple[float, float]]
        ),
        leq: bool = False,
        geq: bool = False,
        eq: bool = False,
        forall: list[_X | _Component] | None = None,
    ):
        self.sample = sample
        self.parameter = parameter
        self.leq = leq
        self.geq = geq
        self.eq = eq
        self.forall = forall

        self.model = sample.model

        self.nominal = sample.nominal
        self.norm = sample.norm
        self.domain = sample.domain
        self.aspect = sample.aspect
        self.report = sample.report
        self.program = sample.program

        # when i say all elements in the set, it implies that each element features
        # individually in the index of the updated sample

        # if as set is passed
        # write the constraint 'for all' elements in it
        if self.forall:

            if isinstance(parameter, list):
                # if a list is passed
                # iterate over it
                for n, idx in enumerate(self.forall):
                    if self.leq:
                        _ = self.sample(idx) <= parameter[n]
                    if self.geq:
                        _ = self.sample(idx) >= parameter[n]
                    if self.eq:
                        _ = self.sample(idx) == parameter[n]
                return

            # if a single value is passed
            # just repeat the same value over
            # all elements in the set

            for idx in self.forall:

                if self.leq:
                    _ = self.sample(idx) <= parameter
                if self.geq:
                    _ = self.sample(idx) >= parameter
                if self.eq:
                    _ = self.sample(idx) == parameter

            return

        if isinstance(parameter, dict):
            # if a dictionary is passed
            # modes are assumed
            n_modes = len(parameter)
            modes_name = f"bin{len(self.model.modes)}"

            setattr(self.model, modes_name, Modes(n_modes=n_modes, bind=self.sample))

            # this gets the last set mode (which was just set above)
            modes = self.model.modes[-1]
            mode_bounds = [
                (
                    (parameter[i - 1], parameter[i])
                    if i - 1 in parameter
                    else (0, parameter[i])
                )
                for i in parameter
            ]
            modes_lb = [b[0] for b in mode_bounds]
            modes_ub = [b[1] for b in mode_bounds]

            _ = self.sample(modes) >= modes_lb

            _ = self.sample(modes) <= modes_ub
            return

        if self.nominal:
            # if a nominal value for the parameter is passed
            # this is essentially the expectation
            # skipping an instance check here
            # if a non iterable is passed, let an error be raised
            if self.norm:
                parameter = normalize(parameter)

            # if the sample needs to be normalized
            parameter = [
                (
                    (self.nominal * i[0], self.nominal * i[1])
                    if isinstance(i, tuple)
                    else self.nominal * i
                )
                for i in parameter
            ]

            # --------- Get LHS
            # lhs needs to be determined here
            # because V will be spaced and timed if not passed by user
            # .X(), .Vb() need time and space

        lhs = self.sample.V(parameter)

        print(f"--- Binding {self.aspect} in domain {self.domain}")

        start = keep_time.time()
        # --------- Get RHS

        if self.aspect.bound is not None:
            # --------- if variable bound
            if self.report:
                # --------- if variable bound and reported
                # we do not want a bi-linear term
                rhs = parameter * self.sample.X(parameter)

            else:
                # --------- if just variable bound
                rhs = parameter * self.sample.Vb()

        elif self.report or self.domain.modes is not None:
            # --------- if  parameter bound and reported or has modes
            # create reporting variable write v <= p*x
            rhs = parameter * self.sample.X(parameter)
            self.aspect.update(self.domain, reporting=True)

        else:
            # --------- if just parameter bound
            rhs = parameter

        if self.leq:
            # Less than equal to
            if (
                self.domain.space in self.aspect.bound_spaces[self.domain.primary]["ub"]
            ) and not self.domain.modes:
                # return if aspect already bound in space
                return
            self.aspect.bound_spaces[self.domain.primary]["ub"].append(
                self.domain.space,
            )
            cons: C = lhs <= rhs
            rel = "_ub"

        elif self.eq:
            # Exactly equal to
            cons: C = lhs == rhs
            rel = "_eq"

        elif self.geq:
            # Greater than equal to
            if (
                self.domain.space in self.aspect.bound_spaces[self.domain.primary]["lb"]
            ) and not self.domain.modes:
                # return if aspect already bound in space
                return
            self.aspect.bound_spaces[self.domain.primary]["lb"].append(
                self.domain.space,
            )
            cons: C = lhs >= rhs
            rel = "_lb"
        else:
            return

        # name of the constraint
        cons_name = rf"{self.aspect.name}{self.domain.idxname}{rel}"

        # categorize the constraint
        if self.domain.modes:
            cons.categorize("Piecewise Linear")
        else:
            cons.categorize("Bound")

        # let all objects in the domain know that
        # a constraint with this name contains it
        self.domain.update_cons(cons_name)

        # let the aspect know about the new constraint
        self.aspect.constraints.append(cons_name)

        # set the constraint
        setattr(
            self.program,
            cons_name,
            cons,
        )

        end = keep_time.time()
        print(f"    Completed in {end-start} seconds")



"""energiapy.commodity init file"""

from .currency import Currency
from .emission import Emission
from .land import Land
from .material import Material
from .resource import Resource

__all__ = [
    "Currency",
    "Emission",
    "Land",
    "Material",
    "Resource",
]



"""Inherited Dimension class"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..represent.model import Model


@dataclass
class _Dimension:
    """
    Inherited by components that are a representation of
    some dimension of the model.

    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, auto generated
    :vartype name: str

    :note:
        - `name` is generated based on the class and model name.
    """

    # model to which the dimension belongs
    model: Model

    def __post_init__(self):
        # There is only one instance of these
        # so they take their own name
        self.name = rf"{self.__class__.__name__}({self.model})"

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __init_subclass__(cls):
        cls.__repr__ = _Dimension.__repr__
        cls.__hash__ = _Dimension.__hash__



"""Data management utilities"""

import numpy as np
import pandas as pd

# from ..solution.result import Result


def make_henry_price_df(
    file_name: str,
    year: int,
    stretch: bool = False,
) -> pd.DataFrame:
    """
    Makes a DataFrame from data with missing values filled using previous day values.

    The costs are converted from $/MMBtu to $/kg using a factor of 1/22.4.
    Only works if there is a full year of data (365 days).
    If `stretch` is True, the timescale is repeated to expand from days (365) to hours (8760).

    Parameters
    ----------
    file_name : :class:`str`
        Path to the CSV file containing the data.
    year : :class:`int`
        Year to import data from.
    stretch : :class:`bool`, optional
        If True, stretches the timescale from days to hours. Defaults to ``False``.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame containing varying natural gas prices with missing values filled.
    """

    df = pd.read_csv(file_name, skiprows=5, names=["date", "CH4"])

    df[["month", "day", "year"]] = df["date"].str.split("/", expand=True)
    df = df[df["year"] == str(year)].astype({"month": int, "day": int, "year": int})
    # , format='%d%b%Y:%H:%M:%S.%f')
    df["date"] = pd.to_datetime(df["date"])
    df["doy"] = df["date"].dt.dayofyear
    df = df.sort_values(by=["doy"])
    df = df.drop(columns="date").dropna(axis="rows")
    doy_list = list(df["doy"])

    # fixes values for weekends and holidays to last active day
    for i in np.arange(1, 366):
        if i not in doy_list:
            if i == 1:  # onetime fix if first day has no value, takes value from day 2
                df = pd.concat(
                    [
                        df,
                        pd.DataFrame.from_records(
                            [
                                {
                                    "CH4": df["CH4"][df["doy"] == 2].values[0],
                                    "month": 1,
                                    "day": 1,
                                    "year": year,
                                    "doy": 1,
                                },
                            ],
                        ),
                    ],
                )
                # df = df.append({'CH4': df['CH4'][df['doy'] == 2].values[0], 'month': 1, 'day': 1, 'year': year, 'doy': 1}, ignore_index=True)
            else:
                df = pd.concat(
                    [
                        df,
                        pd.DataFrame.from_records(
                            [
                                {
                                    "CH4": df["CH4"][df["doy"] == i - 1].values[0],
                                    "month": df["month"][df["doy"] == i - 1].values[0],
                                    "day": df["day"][df["doy"] == i - 1].values[0],
                                    "year": df["year"][df["doy"] == i - 1].values[0],
                                    "doy": i,
                                },
                            ],
                        ),
                    ],
                )
                # df = df.append({'CH4': df['CH4'][df['doy'] == i-1].values[0], 'month': df['month'][df['doy'] == i-1].values[0], 'day': df['day'][df['doy'] == i-1].values[0], 'year': df['year'][df['doy'] == i-1].values[0], 'doy': i}, ignore_index=True)

    df = df.sort_values(by=["doy"])
    df = df.reset_index(drop=True)
    df["CH4"] = df["CH4"] / 22.4  # convert from $/MMBtu to $/kg
    df = df[["CH4", "doy"]].rename(columns={"doy": "day"})
    if stretch is False:
        df = df
        df["scales"] = [(0, int(i - 1)) for i in df["day"]]

    else:
        df = df.loc[df.index.repeat(24)].reset_index(drop=True)
        df["hour"] = [int(i) for i in range(0, 24)] * 365
        df["scales"] = [(0, int(i - 1), int(j)) for i, j in zip(df["day"], df["hour"])]
    df = df[["CH4", "scales"]]
    return df


def remove_outliers(
    data: pd.DataFrame,
    sd_cuttoff: int = 2,
    mean_range: int = 1,
) -> pd.DataFrame:
    """
    Removes outliers up to a chosen number of standard deviations.

    Outliers are replaced with the mean of data points on both sides of the point.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        Input data.
    sd_cutoff : :class:`int`, optional
        Remove data points that are beyond this many standard deviations. Defaults to ``2``.
    mean_range : :class:`int`, optional
        Number of neighboring data points on each side to average over when replacing outliers. Defaults to ``1``.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame with outliers replaced by local means.
    """
    data_mean, data_std = data.mean(), data.std()
    # identify outliers
    cut_off = data_std * sd_cuttoff
    lower, upper = data_mean - cut_off, data_mean + cut_off
    for i in range(len(data)):
        x = data.iloc[i].values[0]
        if x < float(lower.iloc[0]) or x > float(upper.iloc[0]):
            data.iloc[i] = (
                (
                    sum(data.iloc[i - (j + 1)] for j in range(mean_range))
                    + sum(data.iloc[i + (j + 1)] for j in range(mean_range))
                )
                / 2
                * mean_range
            )
    return data


# def get_data(file_name: str) -> dict:
#     """gets data from energia json database

#     Args:
#         file_name (str): name of file

#     Returns:
#         dict: dictionary with data
#     """
#     with open(file_name + '.json') as f:
#         data_ = json.load(f)
#         f.close()
#     return data_


# def dump_data(data: dict, file_name: str):
#     """dump data in any of the following formats: .json, .txt, .pkl

#     Args:
#         data (dict): dictionary with cost trajectories for all processes
#         file_name (str): name of output file with format, e.g. file_name.pkl
#     """

#     if '.pkl' in file_name:
#         with open(file_name, "wb") as f:
#             pickle.dump(data, f)
#             f.close()
#     elif '.json' in file_name:
#         with open(file_name, 'w') as f:
#             json.dump(data, f)

#     elif '.txt' in file_name:
#         with open(file_name, "w") as f:
#             f.write(str(data))
#             f.close()
#     return


# def make_conversion_dict(file_name: str) -> dict:
#     """updates conversion.json conversion values by process

#     Returns:
#         dict: dictionary with conversion values
#     """
#     conversion_dict_ = (
#         pandas.read_csv(file_name, index_col=0)
#         .dropna(axis='rows')
#         .transpose()
#         .to_dict()
#     )
#     dump_data(conversion_dict_, 'conversion.json')
#     return conversion_dict_


# def make_material_dict(file_name: str) -> dict:
#     """updates infra_mat.json which contains infrastructaral material needs by facility

#     Args:
#         file_name (str):name of file

#     Returns:
#         dict: dictionary with infrastructaral material needs
#     """
#     material_dict_ = (
#         pandas.read_csv(file_name, index_col=0)
#         .dropna(axis='rows')
#         .transpose()
#         .to_dict()
#     )
#     dump_data(material_dict_, 'material.json')

#     return material_dict_


# def make_cost_dict(
#     location_list: list, cost_scenario_list: list, process_list: list, year_list: list
# ) -> dict:
#     """intializes an empty dictionary for cost data for all processes

#     Args:
#         location_list (list): list of locations
#         cost_scenario_list (list): list of scenarios
#         process_list (list): list of processes
#         year_list (list): list of years

#     Returns:
#         dict: dictionary with costs
#     """

#     cost_metrics_list = ['CAPEX', 'Fixed O&M', 'Variable O&M', 'units', 'source']
#     cost_dict = {
#         location_.name: {
#             cost_scenario.name: {
#                 process_.name: {
#                     year_: {cost_metric_: {} for cost_metric_ in cost_metrics_list}
#                     for year_ in year_list
#                 }
#                 for process_ in process_list
#             }
#             for cost_scenario in cost_scenario_list
#         }
#         for location_ in location_list
#     }

#     return cost_dict


# def make_f_purchase(
#     location_list: list,
#     day_list: list,
#     hour_list: list,
#     resource_list: list,
#     varying_resource_df: pandas.DataFrame,
# ) -> dict:
#     """makes a dictionary for varying resource costs.
#     minimum resolution hour|

#     Args:
#         location_list (list): list of locations
#         day_list (list): list of days/seasons
#         hour_list (list): list of hours
#         process_list (list): list of processes
#         varying_resource_df (pandas.DataFrame): dataframe with varying resource costs

#     Returns:
#         dict: dictionary containing hourly conversion factors for all processes
#     """

#     f_purchase_dict_ = {
#         location.name: {
#             resource.name: {day: {hour: {} for hour in hour_list} for day in day_list}
#             for resource in resource_list
#         }
#         for location in location_list
#     }
#     for location, resource, day, hour in product(
#         location_list, resource_list, day_list, hour_list
#     ):
#         if resource.name in varying_resource_df:
#             f_purchase_dict_[location.name][resource.name][day][
#                 hour
#             ] = varying_resource_df[varying_resource_df['day'] == day][
#                 resource.name
#             ].values[
#                 0
#             ]  # use day of the year (doy)
#         else:
#             f_purchase_dict_[location.name][resource.name][day][hour] = 1

#     # varying_resource_df2_ = pandas.DataFrame(columns = ['day', 'hour', var])

#     # for day, hour in product(varying_resource_df['day'], hour_list):
#     #     varying_resource_df2_ = varying_resource_df2_.append({'day': day, 'hour': hour, var: varying_resource_df[var][varying_resource_df['day'] == day].values[0] }, ignore_index = True)

#     # , varying_resource_df2_

#     return f_purchase_dict_


# # def make_nrel_cost_df(location: Location, nrel_cost_xlsx, pick_nrel_process_list: list, year_list: list, case: str, crpyears: float) -> pandas.DataFrame:
# #     """makes dataframe for nrel atb data
# #     processes should be specified based on NREL technology tags
# #     classes for technology will be picked based on location
# #     list of cost metrics ('CAPEX', 'Fixed O&M', 'Variable O&M')

# #     Args:
# #         location (location): location object
# #         nrel_cost_xlsx (.xlsx): excel file from NREL ATB
# #         pick_nrel_process_list (list): list of nrel defined processes
# #         year_list (list): list of years
# #         case (str): Market or Research case
# #         crpyears (float): cost recovery period

# #     Returns:
# #         pandas.DataFrame: with nrel costing data for the specified year
# #     """
# #     cost_metrics_list = ['CAPEX', 'Fixed O&M', 'Variable O&M', 'units']

# #     # import nrel atb cost dataset
# #     nrel_cost_df_ = pandas.read_excel(nrel_cost_xlsx, sheet_name='cost')
# #     nrel_cost_df_ = nrel_cost_df_[nrel_cost_df_['technology'].isin(
# #         pick_nrel_process_list)]  # choose technologies
# #     nrel_cost_df_ = nrel_cost_df_[nrel_cost_df_['core_metric_parameter'].isin(
# #         cost_metrics_list)]  # choose costing data to import
# #     nrel_cost_df_ = nrel_cost_df_[nrel_cost_df_[
# #         'core_metric_case'].isin([case])]  # choose market case
# #     nrel_cost_df_ = nrel_cost_df_[nrel_cost_df_['crpyears'].isin(
# #         [crpyears])]  # choose cost recovery period of 20 year
# #     nrel_cost_df_['technology'].replace({'LandbasedWind': 'WF', 'UtilityPV': 'PV', 'Utility-Scale Battery Storage': 'LiI_c',
# #                                         'Pumped Storage Hydropower': 'PSH_c'}, inplace=True)  # replace names to process IDS
# #     nrel_cost_df_ = nrel_cost_df_[((nrel_cost_df_['technology'] == 'PV') & (nrel_cost_df_['techdetail'] == location.PV_class)) |  # class 5 solar PV\
# #                                   ((nrel_cost_df_['technology'] == 'WF') & (nrel_cost_df_[
# #                                    'techdetail'] == location.WF_class))  # class 4 wind farms\
# #                                   | ((nrel_cost_df_['technology'] == 'LiI_c') & (nrel_cost_df_['techdetail'] == location.LiI_class))  # 8hr battery cycle LiI\
# #                                   | ((nrel_cost_df_['technology'] == 'PSH_c') & (nrel_cost_df_['techdetail'] == location.PSH_class))]  # class 3 PSH
# #     year_list = [year_ + 2021 for year_ in year_list]
# #     nrel_cost_df_ = nrel_cost_df_[nrel_cost_df_['core_metric_variable'].isin(
# #         year_list)]  # get data for years in years_list
# #     nrel_cost_df_ = nrel_cost_df_.drop(columns=['index', 'revision', 'atb_year', 'core_metric_key',
# #                                        'core_metric_case', 'crpyears', 'techdetail'])  # drop unnecessary columns
# #     nrel_cost_df_.columns = [
# #         'metric', 'process', 'cost_scenario_list', 'year', 'units', 'cost']  # rename columns
# #     nrel_cost_df_ = nrel_cost_df_.reset_index(drop=True)  # reset index
# #     # change years to int list 0...
# #     nrel_cost_df_['year'] = nrel_cost_df_['year'] - 2021
# #     nrel_cost_df_['cost_scenario_list'] = nrel_cost_df_[
# #         'cost_scenario_list'].str.lower()

# #     # bring all units to $/MW

# #     nrel_cost_df_.loc[nrel_cost_df_.units == '$/KW-yr', ['cost', 'units']] = 1000 * \
# #         nrel_cost_df_.loc[nrel_cost_df_.units ==
# #                           '$/KW-yr']['cost'].values[0], '$/MW'
# #     nrel_cost_df_.loc[nrel_cost_df_.units == '$/kW', ['cost', 'units']] = 1000 * \
# #         nrel_cost_df_.loc[nrel_cost_df_.units ==
# #                           '$/kW']['cost'].values[0], '$/MW'
# #     nrel_cost_df_.loc[nrel_cost_df_.units == '$/MWh', ['cost', 'units']] = 8760 * \
# #         nrel_cost_df_.loc[nrel_cost_df_.units ==
# #                           '$/MWh']['cost'].values[0], '$/MW'

# #     # nrel_cost_df_['cost'][nrel_cost_df_['units']
# #     #                       == '$/KW-yr'] = nrel_cost_df_['cost'][nrel_cost_df_['units'] == '$/KW-yr']*1000
# #     # nrel_cost_df_['units'][nrel_cost_df_['units'] == '$/KW-yr'] = '$/MW'
# #     # nrel_cost_df_['cost'][nrel_cost_df_['units']
# #     #                       == '$/kW'] = nrel_cost_df_['cost'][nrel_cost_df_['units'] == '$/kW']*1000
# #     # nrel_cost_df_['units'][nrel_cost_df_['units'] == '$/kW'] = '$/MW'
# #     # nrel_cost_df_['cost'][nrel_cost_df_['units']
# #     #                       == '$/MWh'] = nrel_cost_df_['cost'][nrel_cost_df_['units'] == '$/MWh']*8760
# #     # nrel_cost_df_['units'][nrel_cost_df_['units'] == '$/MWh'] = '$/MW'

# #     return nrel_cost_df_


# def load_results(filename: str) -> Result:
#     """loads saved results

#     Args:
#         filename (str): file name

#     Returns:
#         Result: a energiapy result type object
#     """
#     file_ = open(filename, 'rb')
#     results_dict = pickle.load(file_)
#     if results_dict['output']['termination'] != 'optimal':
#         print('WARNING: Loading non-optimal results')
#     return Result(
#         name=filename.split('.')[0],
#         output=results_dict['output'],
#         components=results_dict['components'],
#         duals=results_dict['duals'],
#         model_elements=results_dict['model_elements'],
#     )


# def calculate_hourly(
#     data: pandas.DataFrame, column_name: str, what: str = 'mean'
# ) -> pandas.DataFrame:
#     """Finds the mean, min, max for each hour of the year for multi-year data

#     Args:
#         data (pandas.DataFrame): Timeseries data with datetime index
#         column_name (str): name of value column
#         what (str, optional): 'mean', 'max', 'min'. Defaults to 'mean'.

#     Returns:
#         pandas.DataFrame: Output
#     """
#     results = []
#     data_input = copy.deepcopy(data)
#     data_input['datetime'] = data_input.index
#     data_input['month'] = data_input['datetime'].dt.month
#     data_input['day'] = data_input['datetime'].dt.day
#     data_input['hour'] = data_input['datetime'].dt.hour

#     for month in range(1, 13):  # 12 months
#         for day in range(1, 32):  # 31 days (adjust if needed)
#             for hour in range(24):  # 24 hours
#                 target_data_input = data_input[
#                     (data_input['month'] == month)
#                     & (data_input['day'] == day)
#                     & (data_input['hour'] == hour)
#                 ]

#                 if not target_data_input.empty:
#                     if what == 'max':
#                         max_value = target_data_input[column_name].max()
#                         results.append(
#                             {
#                                 'Date': f"{month}/{day}",
#                                 'Hour': hour,
#                                 'Highest Value': max_value,
#                             }
#                         )
#                     if what == 'min':
#                         min_value = target_data_input[column_name].min()
#                         results.append(
#                             {
#                                 'Date': f"{month}/{day}",
#                                 'Hour': hour,
#                                 'Highest Value': min_value,
#                             }
#                         )

#                     if what == 'mean':
#                         mean_value = target_data_input[column_name].mean()
#                         results.append(
#                             {
#                                 'Date': f"{month}/{day}",
#                                 'Hour': hour,
#                                 'Highest Value': mean_value,
#                             }
#                         )
#     results = pandas.DataFrame(results)
#     results = results.drop(columns=['Date', 'Hour'])
#     return results



"""Sample of an Aspect"""

from __future__ import annotations

from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Optional, Self

from gana import I as Idx
from gana import V, inf, sigma, sup

from ...utils.dictionary import merge_trees
from ..constraints.bind import Bind
from ..constraints.calculate import Calculate

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ..._core._component import _Component
    from ..._core._x import _X
    from ..indices.domain import Domain
    from ..variables.aspect import Aspect


# ------------------------------------------------------------------------------
# Sample is the primary constraint generator
# These are essentially a variable waiting to be bound
# ------------------------------------------------------------------------------


class Sample:
    """
    Sets a bound on a variable (V) within a particular domain.

    :param aspect: Aspect to which the constraint is applied.
    :type aspect: Aspect | None
    :param domain: Domain over which the aspect is defined.
    :type domain: Domain | None
    :param timed: If the temporal index is predetermined. Defaults to None.
    :type timed: bool | None
    :param spaced: If the spatial index is predetermined. Defaults to None.
    :type spaced: bool | None

    :ivar name: Name of the bind.
    :vartype name: str | None
    :ivar model: Model to which the generator belongs.
    :vartype model: Model | None
    :ivar program: Gana Program to which the generated constraint belongs.
    :vartype program: Prg | None
    :ivar opr: Operation to bind (in lieu of a single variable). Defaults to None.
    :vartype opr: F | None
    :ivar domains: Set of domains over which the Bind is applied. Defaults to [].
    :vartype domains: Domain
    :ivar hasinc: If the Bind has some incidental calculation. Defaults to False.
    :vartype hasinc: bool

    .. note::
    - ``timed`` and ``spaced`` help skip the calculation of finding the appropriate index.
        For time, this is done based on the length of the input parameter.
    - ``opr`` is useful if providing a combined bound to different variables.
    - ``name`` is generated based on the variable.
    - ``domains`` are updated as the program is built.
    """

    def __init__(
        self,
        aspect: Aspect,
        domain: Domain,
        label: str = "",
        timed: bool = False,
        spaced: bool = False,
    ):

        # this is the aspect for which the constraint is being defined
        self.aspect = aspect
        # the domain is passed when the aspect is called using __call__()
        self.domain = domain

        self.label = label
        # if the temporal index is predetermined
        self.timed = timed
        # if the spatial index is predetermined
        self.spaced = spaced

        self.model = self.aspect.model
        self.program = self.model.program
        self.grb = self.model.grb

        # if the aspect is bound (operate for example)
        self.bound = self.aspect.bound

        # this is set if the aspect needs a reporting binary variable
        self.report: bool = False

        # if incidental calculation is generated
        self.hasinc: bool = False

        # if nominal is provided
        # and multiplied by the nominal value
        self.nominal: Optional[float] = None
        # the input argument is normalized if True
        self.norm: bool = False

        # the bound is set for all indices
        self._forall: list[_X] = []

    @property
    def name(self) -> str:
        """Name of the constraint"""
        return f"{self.domain.primary}.{self.aspect.name}"

    @property
    def index(self) -> list[_Component]:
        """_Index"""
        return self.domain.index

    @property
    def index_short(self) -> list[_X | Sample]:
        """Short Index"""
        return self.domain.index_short

    @cached_property
    def I(self) -> Idx:
        """gana index set (I)"""
        return self.aspect.I

    @property
    def x(self) -> Self:
        """Binary Reporting Variable"""
        self.report = True
        return self

    @property
    def add(self) -> Self:
        """Positive Control Response"""
        return self.aspect.add(*self.index)

    @property
    def sub(self) -> Self:
        """Negative Control Response"""
        return self.aspect.sub(*self.index)

    @property
    def F(self):
        """Function"""
        return self.V(1)

    @property
    def constraints(self):
        """Constraints"""
        return self.aspect.constraints

    def show(self, descriptive=False):
        """Pretty print the component"""
        constraints = list(chain.from_iterable(i.constraints for i in self.index))

        for c in constraints:
            if c in self.aspect.constraints:
                cons: C = getattr(self.program, c)
                cons.show(descriptive)

    def V(
        self,
        parameters: Optional[float | list] = None,
        length: Optional[int] = None,
        report: bool = False,
        incidental: bool = False,
    ) -> V:
        """Returns a gana variable (V) using .domain as the index.
        If time and space are (or) not given, i.e. .spaced or .timed are False,
        They can be determined.

        For identifying the appropriate temporal scale (index):
            The length of the parameter set is used
            Alternatively, the length of the parameter set can passed directly
            If neither are passed, it defaults to horizon

        If the spatial index is not given, it defaults to the network.


        Args:
            parameters (int | list): the parameter/parameter set. Defaults to None.
            length (int): length of the parameter set. Defaults to None.
            report (bool): to make a binary reporting variable. Defaults to False.
            incidental (bool): if this is an incidental calculation. Defaults to False.
        Note:
            - parameters and length are mutually exclusive
        """

        # with lag it is assumed that the variable of which this is a lagged subset is
        # already defined
        # for example, if opr_t = opr_t-1 + x_t, then opr_t is already defined
        if self.domain.lag:
            # t - 1 is made after t, so no need to set a new one
            try:
                return getattr(self.program, self.aspect.name)(*self.domain.Ilist)
            except KeyError:
                # the variable has not been defined yet
                lag = self.domain.lag
                self.domain = self.domain.change(
                    {"lag": None, "periods": self.domain.lag.of},
                )
                self.V(
                    parameters=parameters,
                    length=length,
                    report=report,
                    incidental=incidental,
                )
                self.domain = self.domain.change({"lag": lag, "periods": None})
                return getattr(self.program, self.aspect.name)(*self.domain.Ilist)

        # ------ Check time and space -------
        # this is only called if the bind variable has no temporal index defined
        def time():
            """Matches an appropriate temporal scale"""
            if isinstance(parameters, list):
                # if list is given, find using length of the list
                if self.domain.modes is not None:
                    self.domain.periods = self.aspect.time.find(
                        len(parameters) / len(self.domain.modes),
                    )
                else:
                    self.domain.periods = self.aspect.time.find(len(parameters))

            elif isinstance(length, int):
                # if length is given, use it directly
                self.domain.periods = self.aspect.time.find(length)
            else:
                # else the size of parameter set is exactly one
                # or nothing is given, meaning the variable is not time dependent
                # thus, index by horizon
                self.domain.periods = self.aspect.horizon

        # this is only called if the bind variable has no spatial index defined
        def space():
            """Assigns to network is spatial index is not given"""
            # if spatial index is not explicity given
            # default to the network
            self.domain.location = self.aspect.network

        if not self.spaced:
            # if the spatial index is not passed
            space()

        if self.domain.primary not in self.aspect.bound_spaces:
            self.aspect.bound_spaces[self.domain.primary] = {"ub": [], "lb": []}

        if not self.timed:
            # if the temporal index is not passed
            time()

        # get the list of Indices in the domain
        index = tuple(self.domain.Ilist)

        # --------- if reporting binary ---------------
        # if a reporting binary variable is needed
        if report:
            # these are basically named using a breve over the variable name or latex name

            if self.aspect.latex:
                ltx = r"{\breve{" + self.aspect.latex + r"}}"
            else:
                ltx = r"{\breve{" + self.aspect.name + r"}}"
            # create a binary variable
            setattr(
                self.program,
                f"x_{self.aspect.name}",
                V(
                    *index,
                    mutable=True,
                    ltx=ltx,
                    bnr=True,
                ),
            )
            v_rpt = getattr(self.program, f"x_{self.aspect.name}")
            self.aspect.reporting = v_rpt
            return v_rpt(*index)

        # if incidental calculation is needed
        # incidental calculations do not scale with variable value
        # rather, they are incurred if the reporting binary = 1
        # see the equations below:
        #   calc_total = calc + calc_incidental
        #   calc = v * param
        #   calc_incidental = v_reporting * param_incidental

        # --------- if incidental ---------------

        if incidental:
            # named with a superscript inc
            if self.aspect.latex:
                ltx = self.aspect.latex + r"^{inc}"
            else:
                ltx = self.aspect.name + r"^{inc}"

            # create an incidental variable (continuous)
            setattr(
                self.program,
                f"{self.aspect.name}_incidental",
                V(*index, mutable=True, ltx=ltx),
            )
            return getattr(self.program, f"{self.aspect.name}_incidental")(*index)

        # --------- if continuous ---------------

        # the reason we check by name:
        # some variables can serve as indices, a normal check ends by
        # creating a constraint variable == variable
        # if not [i.name for i in index] in [
        #     [i.name for i in idx] for idx in self.aspect.indices
        # ]:

        if [i.name for i in index] not in [
            [i.name for i in idx] for idx in self.aspect.indices
        ]:
            # if not index in self.aspect.indices:

            # if a variable has not been created for the index
            # create a variable
            # all energia variables are mutable by default
            setattr(
                self.program,
                self.aspect.name,
                V(*index, mutable=True, ltx=self.aspect.latex),
            )

            # updated the indices
            # we can be confident that the index is unique
            # because of the check above
            self.aspect.indices.append(index)

            # this updates the balanced dictionary, by adding the commodity as a key

            if self.domain.commodity and not self.domain.lag:
                _ = self.grb[self.domain.commodity][self.domain.space][self.domain.time]

            # this lets all index elements in the domain know
            # that the aspect was sampled
            self.domain.update_domains(self.aspect)

            # --------- Update the disposition ---------------

            # get the primary component
            # update the disposition dictionary
            self.model.dispositions = merge_trees(
                self.model.dispositions,
                {self.aspect: self.domain.tree},
            )

            # for the same aspect, map variables with higher order indices
            # to variables with lower order indices
            self.aspect.update(self.domain)

            self.aspect.domains.append(self.domain)

        return getattr(self.program, self.aspect.name)(*index)

    def Vinc(self, parameters: float | list = None, length: int = None) -> V:
        """Returns the incidental variable

        Args:
            parameters (float | list): the parameter/parameter set. Defaults to None.
            length (int): length of the parameter set. Defaults to None.

        Returns:
            V: the incidental variable
        """
        self.hasinc = True
        return self.V(parameters, length, incidental=True)

    def Vb(self) -> V:
        r"""Bound Variable

        Returns:
            V: the bound variable

        These apply when there are multiple levels of variable-making
        Endogenous bounds apply, i.e.

        ..math::
           \mathbf{v}_{\dots, t^{+}} <= {\theta}_{\dots, t^{+}} \cdot \mathbf{v}_{\dots, t^{-}}

        where :
        - :math:`\mathbf{v}` is the variable
        - :math:`\theta` is the parameter set
        - :math:`t^{i}` are bespoke discretizations of the horizon
        - :math:`i \in \{-, +\}` are variable-making levels
        """

        bound_aspect = getattr(self.model, self.aspect.bound)

        if bound_aspect not in self.model.dispositions:
            return 1
            # print(
            #     f"--- Aspect ({bound_aspect}) not defined, a variable will be created at {self.domain.space} assuming {self.model.horizon} as the temporal index",
            # )

            # _ = bound_aspect(self.domain.primary, self.domain.space).V()

        if (
            self.domain.space
            not in self.model.dispositions[bound_aspect][self.domain.primary]
        ):

            # if the bound variable has not been defined at the given space
            print(
                f"--- Aspect ({bound_aspect}) not defined at {self.domain.space}, a variable will be created assuming {self.model.horizon} as the temporal index",
            )

            domain = self.domain.change({"periods": self.model.horizon})

        else:
            # if the bound variable has been defined for the given space
            times = self.model.dispositions[bound_aspect][self.domain.primary][
                self.domain.space
            ]
            time = max(list(times))
            if time >= self.domain.periods:
                domain = self.domain.change({"periods": time})
            else:
                # this is if the binding variable has a sparser temporal index compared to time
                raise ValueError(
                    f"Incompatible temporal indices: {bound_aspect} defined in time {time} and {self.aspect} defined in time {self.domain.periods}\n"
                    f"Binding variable ({bound_aspect}) cannot have a denser discretization than variable being bound ({self.aspect})",
                )
        return bound_aspect(domain=domain).V()

    def X(self, parameters: float | list = None, length: int = None) -> V:
        r"""Binary Reporting Variable

        These report whether a variable has been made or not
        Also useful to make the variable space semi-continuous

        ..math::
              \mathbf{v}_{\dots, t} <= {\theta}_{\dots, t} \cdot \mathbf{x}_{\dots, t}

        where :
            - :math:`\mathbf{v}` is the variable
            - :math:`\mathbf{x}` is the binary reporting variable
            - :math:`\theta` is the parameter set
            - :math:`t` some bespoke discretization of the horizon
        """
        return self.V(parameters, length, report=True)

    def obj(self, max: bool = False):
        """Set the sample itself as the objective

        max (bool): if maximization, defaults to False
        """
        if not self.timed:
            # if the temporal index is not passed
            self.domain.periods = self.model.horizon
        if not self.spaced:
            # if the spatial index is not passed
            self.domain.location = self.model.network

        # consider all of self.domain
        v = self.V()

        if len(v) == 1:
            _obj = v
        else:

            _obj = sigma(v)

        if self.hasinc:
            # if there is an incidental variable
            # the incidental variable is added to the objective
            v_inc = self.Vinc()
            if len(v_inc) == 1:
                _obj += v_inc
            else:
                _obj += sigma(v_inc)

        if max:
            setattr(self.program, f"max{self.aspect.name})", sup(_obj))
        else:
            setattr(self.program, f"min({self.aspect.name})", inf(_obj))

        self.program.renumber()

    def opt(self, max: bool = False):
        """Optimize

        max (bool): if maximization, defaults to False
        """
        self.obj(max)
        # optimize!
        self.program.opt()

    def bounds(self):
        """Finds the bounds of the variable"""
        # TODO
        self.opt()
        bmin = self.program.obj()
        self.opt(max=True)
        bmax = -self.program.obj()
        return (bmin, bmax)

    def prep(self, nominal: float = 1, norm: bool = True) -> Self:
        """Nominal value
        Args:
            value (float): Nominal value to multiply with bounds
            norm (bool): If the input argument (bounds) are normalized, defaults to True
        """
        self.nominal = nominal
        self.norm = norm
        return self

    def sol(self, aslist: bool = False):
        """Solution

        Args:
            aslist (bool, optional): Returns the solution as list, otherwise as a variable
        """
        return self.V().sol(aslist=aslist)

    def eval(self, *values: float):
        """Evaluate the variable using parametric variable values
        Args:
            *values (float): values for the parametric variables
        """
        return self.V().eval(*values)

    def forall(self, index) -> Self:
        """Returns the function at the given index"""
        self._forall = index
        return self

    def __le__(self, other):

        Bind(sample=self, parameter=other, leq=True, forall=self._forall)

    def __ge__(self, other):
        Bind(sample=self, parameter=other, geq=True, forall=self._forall)

    def __eq__(self, other):

        if other is True:
            # if a truth value is give
            # just declare the variable
            # it will be non-negative by default
            # and will begin a commodity balance
            self.V()

        elif isinstance(other, Sample):
            if self.aspect.name == other.aspect.name:
                # if self.domain == other.domain:
                return True
            return False

        else:
            Bind(sample=self, parameter=other, eq=True, forall=self._forall)

    def __gt__(self, other):
        print(f"--- Bind {self} > {other} is being written as {self} >= {other}")
        _ = self >= other

    def __lt__(self, other):
        print(f"--- Bind {self} < {other} is being written as {self} <= {other}")
        _ = self <= other

    def __call__(self, *index) -> Self:

        index = list(set(self.domain.index_short + list(index)))
        sample = self.aspect(*index)
        sample.report = self.report
        return sample

    def __getitem__(self, dependent: Sample):
        if isinstance(dependent, int):
            f = self.F(self.F.index[dependent])
            f.report = self.report
            return f

        calculation = dependent(self)
        decision = self(*self.index_short)
        decision.report = self.report

        return Calculate(calculation=calculation, sample=decision)

        # variable.report = self.report
        # return Calculate(calc=dependent(*self.index), decision=variable)

    def draw(self, **kwargs):
        """Draws the variable"""
        v = self.V()
        v.draw(**kwargs)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)


# This is for binding sample operations, eg. sample + sample or sample - sample


#    def __add__(self, other: Self | FBind):
#         if isinstance(other, (int, float)):
#             return FBind(F=self.F + other, program=self.program)
#         return FBind(F=self.F + other.F, program=self.program)

#     def __radd__(self, other):
#         if not other:
#             return self

#     def __sub__(self, other: Self | FBind):
#         if isinstance(other, (int, float)):
#             return FBind(F=self.F - other, program=self.program)
#         return FBind(F=self.F - other.F, program=self.program)

#     def __rsub__(self, other: int | float):
#         return FBind(F=other - self.F, program=self.program)

#     def __mul__(self, other: Self | FBind):
#         return FBind(F=self.F * other.F, program=self.program)

#     def __rmul__(self, other: int | float):
#         return FBind(F=other * self.F, program=self.program)

# class FBind:
#     """Function Bind

#     This is used to bind a function of variables to a given parameter (set)

#     """

#     def __init__(self, F: F, program: Prg):
#         self.program = program
#         self.F = F

#     def __add__(self, other: Self | Sample):
#         if isinstance(other, (int, float)):
#             return FBind(F=self.F + other, program=self.program)
#         return FBind(F=self.F + other.F, program=self.program)

#     def __radd__(self, other):
#         if not other:
#             return self

#     def __sub__(self, other: Self | Sample):
#         if isinstance(other, (int, float)):
#             return FBind(F=self.F - other, program=self.program)
#         return FBind(F=self.F - other.F, program=self.program)

#     def __rsub__(self, other: int | float):
#         return FBind(F=other - self.F, program=self.program)

#     def __mul__(self, other: Self | Sample):
#         return FBind(F=self.F * other.F, program=self.program)

#     def __rmul__(self, other: int | float):
#         return FBind(F=other * self.F, program=self.program)

#     def __eq__(self, other):
#         setattr(self.program, f"eq_{self.F.name}", self.F == other)

#     def __le__(self, other):
#         setattr(self.program, f"le_{self.F.name}", self.F <= other)

#     def __ge__(self, other):
#         setattr(self.program, f"ge_{self.F.name}", self.F >= other)



"""Utilities to perform mathematical operations"""

from math import erf, exp, pi, sqrt

import numpy


def norm_constant(p, mu, sigma) -> float:
    """Calculates the normal constant

    Args:
        p (float): level of complaince
        mu (float): mean
        sigma (float): standard deviation

    Returns:
        float: normal constant
    """
    x = mu + erf(1 / sqrt(2) * p) * sigma * sqrt(2)
    return 1 / (sigma * sqrt(2 * pi)) * exp(-((x - mu) ** 2) / (2 * sigma**2))


def find_euclidean_distance(cluster_node_a: list, cluster_node_b: list) -> float:
    """finds euclidean distances between two cluster nodes

    Args:
        cluster_node_a (float): index tag for cluster node a
        cluster_node_b (float): index tag for cluster node b

    Returns:
        float: euclidean distance
    """
    euclidean_distance_ = [(a - b) ** 2 for a, b in zip(cluster_node_a, cluster_node_b)]
    euclidean_distance_ = sum(euclidean_distance_)
    return euclidean_distance_


def generate_connectivity_matrix(scale_len) -> numpy.array:
    """generates a connectivity matrixto maintain chronology [..1,0,1..]

    Returns:
        numpy.array: matrix with connectivity relations
    """
    connect_ = numpy.zeros((scale_len, scale_len), dtype=int)
    for i_ in range(len(connect_)):

        if i_ == 0:
            # connect_[i,364] = 1 #uncomment these to generate a cyclic matrix
            connect_[i_, 1] = 1
        elif i_ == scale_len - 1:
            connect_[i_, scale_len - 2] = 1
            # connect_[i,0] = 1
        else:
            connect_[i_, i_ - 1] = 1
            connect_[i_, i_ + 1] = 1
    return connect_


def normalize(data: list, how: str = "max") -> list:
    """min max normalization of data

    Args:
        data (numpy.array | DataFrame): time-series data
        how (str): min-max or max
    Returns:
        numpy.array | DataFrame: min-maxed data array
    """
    if all(isinstance(i, tuple) for i in data):
        # lower bound
        lb = [i[0] for i in data]
        # upper bound
        ub = [i[1] for i in data]
        # normalize lower and upper bounds individually
        lb_norm = normalize(lb, how=how)
        ub_norm = normalize(ub, how=how)
        return list(zip(lb_norm, ub_norm))

    if how == "min_max":
        min_ = min(data)
        max_ = max(data)
        data = [(i - min_) / (max_ - min_) for i in data]

    if how == "max":
        max_ = max(data)
        data = [i / max(data) for i in data]
    return data



"""Init library examples."""

from .energy import (
    design_scheduling,
    design_scheduling_materials,
    scheduling,
    supermarket,
)

__all__ = [
    "design_scheduling",
    "design_scheduling_materials",
    "scheduling",
    "supermarket",
]



"""Time"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING
from warnings import warn

from .._core._dimension import _Dimension
from ..components.temporal.modes import Modes
from ..components.temporal.periods import Periods

if TYPE_CHECKING:
    from gana.block.program import Prg


@dataclass
class Time(_Dimension):
    """
    Temporal representation of a system.

    All time periods are attached to this object.

    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, generated based on the class and model name.
    :vartype name: str
    :ivar periods: List of time periods. Defaults to [].
    :vartype periods: list[Periods]
    :ivar modes: List of modes. Defaults to [].
    :vartype modes: list[Modes]

    .. note::
    - name is generated based on the Class and Model name
    - periods are populated as the program is built.
    - _indexed is made the first time Time is indexed.
    """

    def __post_init__(self):
        self.periods: list[Periods] = []
        self.modes: list[Modes] = []
        _Dimension.__post_init__(self)

    @cached_property
    def program(self) -> Prg:
        """Mathematical program"""
        return self.model.program

    # -----------------------------------------------------
    #                    Helpers
    # -----------------------------------------------------

    @property
    def tree(self) -> list[Periods]:
        """Return the tree of periods"""
        hrz = self.horizon

        return {int(hrz.howmany(prd)): prd for prd in self.periods}

    @property
    def sorted_periods(self) -> list[Periods]:
        """Sorted periods from densest to sparsest"""
        return sorted(self.periods)

    def find(self, size: int | float) -> Periods:
        """Find the period that has the length"""

        if size not in self.tree:
            # if no math make a default period
            warn(
                f"{size} does not match the size of any data set passed, generating 't{size}'.)",
            )
            _ = self.model.default_periods(size=size)

        return self.tree[size]

    def split(self, period: Periods) -> tuple[list[Periods], list[Periods]]:
        """Gives a list of periods which are denser and sparser than period"""

        periods = self.sorted_periods
        index = periods.index(period)
        return periods[:index], periods[index + 1 :]

    # -----------------------------------------------------
    #                    Superlatives
    # -----------------------------------------------------

    @property
    def densest(self) -> Periods:
        """The densest period"""
        if self.periods:
            return min(self.periods, key=lambda x: x.periods)
        return self.horizon

    @property
    def sparsest(self) -> Periods:
        """The sparsest period"""
        if self.periods:
            return max(self.periods, key=lambda x: x.periods)
        return self.horizon

    @property
    def horizon(self) -> Periods:
        """The sparsest scale is treated as the horizon"""
        if self.periods:
            return self.sparsest
        # if nothing found, make a default period
        return self.model.default_periods()



"""Utilities for plots, manages axes and such"""

from matplotlib import ticker


def axis_formatter(axes, xcord: list, axis_labels: str = "M"):
    pos_list_8760 = [
        0,
        744,
        1344,
        2160,
        2800,
        3624,
        4344,
        5088,
        5832,
        6552,
        7296,
        8016,
    ]  # hours of the year corresponding to month
    pos_list_365 = [
        0,
        31,
        59,
        90,
        120,
        151,
        181,
        212,
        243,
        273,
        304,
        334,
    ]  # hours of the year corresponding to month
    name_list = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]
    if axis_labels == "M" and len(xcord) == 8760:
        axes.xaxis.set_major_locator(ticker.FixedLocator((pos_list_8760)))
        axes.xaxis.set_major_formatter(ticker.FixedFormatter((name_list)))
    if axis_labels == "M" and len(xcord) == 365:
        axes.xaxis.set_major_locator(ticker.FixedLocator((pos_list_365)))
        axes.xaxis.set_major_formatter(ticker.FixedFormatter((name_list)))
    if axis_labels == "M" and len(xcord) == 12:
        axes.xaxis.set_major_locator(ticker.FixedLocator((list(range(12)))))
        axes.xaxis.set_major_formatter(ticker.FixedFormatter((name_list)))

    else:
        pass
    return axes



"""Domain"""

from __future__ import annotations

from dataclasses import dataclass, field
from math import prod
from operator import is_, is_not
from typing import TYPE_CHECKING, Optional, Self

if TYPE_CHECKING:
    from gana import I as Idx
    from gana import V

    from ..._core._x import _X
    from ...components.commodity._commodity import _Commodity
    from ...components.game.couple import Couple
    from ...components.game.player import Player
    from ...components.impact.indicator import Indicator
    from ...components.operation.process import Process
    from ...components.operation.storage import Storage
    from ...components.operation.transport import Transport
    from ...components.spatial.linkage import Linkage
    from ...components.spatial.location import Location
    from ...components.temporal.lag import Lag
    from ...components.temporal.modes import Modes
    from ...components.temporal.periods import Periods
    from ...represent.model import Model
    from ..variables.aspect import Aspect
    from ..variables.sample import Sample


@dataclass
class Domain:
    """
    A domain is an ordered set of the indices of an element.

    :param indicator: Indicates the impact of some activity through an equivalency,
        e.g. GWP, ODP.
    :type indicator: Indicator | None

    :param commodity: Represents the flow of any stream, measured using some basis,
        e.g. water, Rupee, carbon-dioxide.
    :type commodity: _Commodity | None

    :param process: Process that is being considered, e.g. dam, farming.
    :type process: Process | None

    :param storage: Storage that is being considered, e.g. reservoir.
    :type storage: Storage | None

    :param transport: Transport that is being considered, e.g. pipeline, road.
    :type transport: Transport | None

    :param player: Actor that takes decisions, e.g. me, you.
    :type player: Player | None

    :param couple: Other actor that might be paired with the player.
    :type couple: Couple | None

    :param location: Spatial aspect of the domain, e.g. Goa, Texas.
    :type location: Location | None

    :param linkage: Linkage aspect of the domain, e.g. pipeline, road.
    :type linkage: Linkage | None

    :param periods: Temporal aspect of the domain, e.g. year, month.
    :type periods: Periods | None

    :param lag: Indicates whether the temporal element is lagged or not.
    :type lag: Lag | None

    :param modes: Modes applicable to the domain.
    :type modes: Modes | None

    :param binds: List of binds that can be summed over.
    :type binds: list[Bind] | None

    :ivar model: Model to which the Domain belongs.
    :vartype model: Model

    """

    # the reason I keep these individual
    # instead of directly adding primary or stream
    # is because we do an instance check in Aspect
    # this helps relay the checks
    # primary component (one of these is needed)
    indicator: Optional[Indicator] = None
    commodity: Optional[_Commodity] = None

    process: Optional[Process] = None
    storage: Optional[Storage] = None
    transport: Optional[Transport] = None

    # decision - maker and other decision-maker
    player: Optional[Player] = None
    couple: Optional[Couple] = None

    # compulsory space and time elements
    location: Optional[Location] = None
    linkage: Optional[Linkage] = None
    periods: Optional[Periods] = None
    lag: Optional[Lag] = None
    modes: Optional[Modes] = None

    # These can be summed over
    binds: Optional[list[Sample]] = field(default_factory=list)

    def __post_init__(self):
        # Domains are structured something like this:
        # (primary_component ...aspect_n, secondary_component_n....,decision-makers, space, time
        # primary_component can be an indicator, commodity, or operation (process | storage | transport)
        # {aspect_n: secondary_component_n} is given in self.binds
        # decision-makers = player | couple
        # space = location | linkage
        # time = period | lag

        # primary index being modeled in some spatiotemporal context
        self.model: Model = next((i.model for i in self.index_short if i), None)

    @property
    def I(self) -> tuple[Idx, ...]:
        """Compound index"""
        return prod(self.Ilist)

    # -----------------------------------------------------
    #                    Components
    # -----------------------------------------------------

    # These are kept as properties
    # because domains are updated on the fly
    # look at change() and call()

    @property
    def stream(self) -> Indicator | _Commodity:
        """Stream"""
        return self.indicator or self.commodity

    @property
    def operation(self) -> Process | Storage | Transport:
        """Operation"""
        return self.process or self.storage or self.transport

    @property
    def primary(self) -> Indicator | _Commodity | Process | Storage | Transport:
        """Primary component"""
        _primary = self.stream or self.operation
        if not _primary:
            raise ValueError("Domain must have at least one primary index")
        return _primary

    @property
    def space(self) -> Location | Linkage:
        """Space"""
        return self.linkage or self.location

    @property
    def maker(self) -> Player | Couple:
        """Decision-maker"""
        return self.couple or self.player

    @property
    def time(self) -> Periods | Lag:
        """Time"""
        if self.periods is not None:
            return self.periods
        if self.lag is not None:
            return self.lag

        return self.model.horizon

    @property
    def linked(self) -> bool:
        """Linked"""
        return True if self.linkage else False

    @property
    def lagged(self) -> bool:
        """Lagged"""
        return True if self.lag else False

    # -----------------------------------------------------
    #                    Disposition
    # -----------------------------------------------------

    @property
    def isroot(self) -> bool:
        """This implies that the domain is of the form
        <object, space, time>
        """
        if self.lag:
            return False
        if self.disposition in [
            ("commodity", "space", "time"),
            ("indicator", "space", "time"),
            ("operation", "space", "time"),
        ]:
            return True
        return False

    @property
    def isrootroot(self) -> bool:
        """This implies that the domain is of the form
        <object, network, horizon>
        Thus, an element attached to this domain has the
        lowest possible dimensionality
        """
        if self.isroot:
            if is_(self.time, self.time.horizon):
                if is_(self.space, self.space.network):
                    return True
        return False

    @property
    def disposition(self) -> tuple[str, ...]:
        """Disposition"""
        return tuple(self._.keys())

    # -----------------------------------------------------
    #                    Naming
    # -----------------------------------------------------

    @property
    def name(self):
        """Name"""
        return f"{tuple(self.index)}"

    @property
    def idxname(self):
        """Name of the index"""
        return "_" + "_".join(f"{i}" for i in self.index)

    # -----------------------------------------------------
    #                    Dictionaries
    # -----------------------------------------------------

    @property
    def index(self) -> list[Aspect | _X]:
        """list of _Index elements"""

        # binds = sum([[i, j] for i, j in self.binds.items()], [])

        # these default to network and horizon, so can default from model using .space or .time

        return self.index_primary + self.index_binds

    @property
    def index_primary(
        self,
    ) -> list[Indicator | _Commodity | Process | Storage | Transport]:
        """Primary index

        Returns:
            list[X]: list of primary indices
        """
        return [self.primary] + [
            i
            for i in [self.location, self.linkage, self.periods, self.lag, self.modes]
            if i is not None
        ]

    @property
    def index_binds(self) -> list[Aspect | _X]:
        """List of bind indices

        Returns:
            list[Bind, X]: list of bind indices
        """
        return [x for b in self.binds for x in (b.aspect, b.domain.primary)]

    @property
    def index_short(
        self,
    ) -> list[Indicator | _Commodity | Process | Storage | Transport | Sample]:
        """Set of indices"""
        return self.index_primary + self.binds

    @property
    def tree(self) -> dict:
        """Convert index into tree"""

        tree = {}
        node = tree

        if self.binds:
            index = self.index[: -2 * (len(self.binds))]
        else:
            index = self.index
        for key in index:
            node[key] = {}
            node = node[key]
        for b in self.binds:
            node[b.aspect] = {}
            node[b.aspect][b.domain.primary] = {}
            node = node[b.aspect][b.domain.primary]

        # tree = {}
        # node = tree

        return tree

    @property
    def aspects(self) -> list[Aspect]:
        """Aspects"""
        return [b.aspect for b in self.binds]

    @property
    def args(self) -> dict[str, _X | Lag | Modes | list[Sample]]:
        """Dictionary of indices"""
        return {
            "indicator": self.indicator,
            "commodity": self.commodity,
            "player": self.player,
            "process": self.process,
            "storage": self.storage,
            "transport": self.transport,
            "location": self.location,
            "linkage": self.linkage,
            "periods": self.periods,
            "lag": self.lag,
            "modes": self.modes,
            "binds": self.binds,
        }

    @property
    def dictionary(self) -> dict[str, _X | Lag | Modes | list[Sample]]:
        """Dictionary of indices"""
        return {
            "primary": self.primary,
            "player": self.player,
            "space": self.space,
            "time": self.time,
            "modes": self.modes,
            "binds": self.binds,
        }

    # -----------------------------------------------------
    #                    Iterables
    # -----------------------------------------------------

    @property
    def _(self):
        """Dictionary of indices that are not None"""
        return {i: j for i, j in self.dictionary.items() if j is not None}

    @property
    def tup(self):
        """Tuple of objects"""
        return tuple(self._.values())

    @property
    def lst(self):
        """Tuple of objects"""
        return list(self._.values())

    @property
    def Ilist(self) -> list[Idx | list[V]]:
        """List of I"""
        return [i if isinstance(i, list) else i.I for i in self.index]

    # -----------------------------------------------------
    #                    Helpers
    # -----------------------------------------------------

    def update_cons(self, cons_name: str):
        """Update the constraints declared at every index"""
        for j in self.index:
            if cons_name not in j.constraints:
                j.constraints.append(cons_name)

    def update_domains(self, aspect: Aspect):
        """Update all elements in the domains with the aspects
        that they have been modeled in
        """
        for i, j in self._.items():
            if self.lag and i == "time":
                # lags disappear anyway, so dont bother
                continue
            if i not in ["binds"]:
                # these are dependent variables, so do not update them
                if self not in j.domains:
                    # check and update the domains at each index
                    j.domains.append(self)
                # if the variable is not in the list of variables at the index
                # update those
                if aspect not in j.aspects:
                    # first time (variable) is a dict {aspect: [..aspects..]}
                    j.aspects[aspect] = [self]
                elif self not in j.aspects[aspect]:
                    j.aspects[aspect].append(self)

    def copy(self) -> Self:
        """Make a copy of self"""
        return Domain(**self.args)

    def change(self, what: dict[str, _X]) -> Self:
        """Change some aspects and return a new Domain"""
        return Domain(**{**self.args, **what})

    # -----------------------------------------------------
    #                    Vector
    # -----------------------------------------------------

    def __getitem__(self, index: str) -> _X:
        """Get the index by name"""
        return self._[index]

    def __iter__(self):
        """Iterate over the indices"""
        return iter(self.index_short)

    def __call__(self, *args: str) -> Self:
        return Domain(**{i: j for i, j in self.args.items() if i in args})

    def __len__(self):
        return len(self.disposition)

    # -----------------------------------------------------
    #                    Operators
    # -----------------------------------------------------

    def __truediv__(self, other: list[str]) -> Self:
        """Will give you the Domain minus a particular index

        Args:
            other (str): index you wish to remove

        Returns:
            Self: lower dimensional domain

        Example:
            >>> domain = Domain(commodity=water, operation=dam, space=goa, time=year)
            (water, dam, goa, year)
            >>> domain/'operation'
            (water, goa, year)
            >>> domain/['commodity', 'operation']
            (goa, year)
        """

        return Domain(**{i: j for i, j in self.args.items() if i not in other})

    def __sub__(self, other: Self) -> list[str]:
        """Will give you a list of indices that are not common between two domains
        Args:
            other (Self): another Domain object
        Returns:
            list[str]: list of indices that are not common

        Example:
            >>> domain1 = Domain(commodity=water, operation=dam, space=goa, time=year)
            >>> domain2 = Domain(commodity=water, space=mumbai, time=year)
            >>> domain1 - domain2
            ['operation', 'space']

        """
        notcommon = []
        for i, j in self._.items():
            # append if disposition is same but value is not
            if i in other._:
                if is_not(other._[i], j):
                    notcommon.append(i)
            else:
                # append if disposition is not in other
                notcommon.append(i)
        # The first loop will not check for the keys that are not in self._
        for k in other._.keys():
            if k not in self._:
                notcommon.append(k)

        return notcommon

    def __add__(self, other: Self) -> Self:
        return Domain(**{**self.args, **other.args})

    # -----------------------------------------------------
    #                    Relational
    # -----------------------------------------------------

    def __eq__(self, other):
        return self.name == str(other)

    def __lt__(self, other: Self) -> bool:
        """Less than comparison based on the number of indices"""
        if len(self) < len(other):
            return True
        return False

    def __gt__(self, other: Self) -> bool:
        """Greater than comparison based on the number of indices"""
        return other < self

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



"""Recipes to create Aspects"""

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Type

from .aspect import Aspect

if TYPE_CHECKING:
    from ..._core._component import _Component


@dataclass
class Recipe:
    """Recipe to create Aspects"""

    name: str
    kind: Type[Aspect]
    primary_type: tuple[Type[_Component]] | Type[_Component]
    label: str = ""
    add: str = ""
    sub: str = ""
    bound: str = ""
    ctrl: str = ""
    ispos: bool = True
    nn: bool = True
    latex: str = ""

    def __post_init__(self):
        self.args = {
            k: v
            for k, v in asdict(self).items()
            if k not in ("name", "kind") and (isinstance(v, bool) or v)
        }

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



"""Energia Imports"""

from .components.commodity.currency import Currency
from .components.commodity.emission import Emission
from .components.commodity.land import Land
from .components.commodity.material import Material
from .components.commodity.resource import Resource
from .components.game.player import Player
from .components.impact.categories import Economic, Environ, Social
from .components.measure.unit import Unit
from .components.operation.process import Process
from .components.operation.storage import Storage
from .components.operation.transport import Transport
from .components.spatial.linkage import Linkage
from .components.spatial.location import Location
from .components.temporal.periods import Periods
from .components.temporal.scales import TemporalScales
from .library.components import (
    currencies,
    env_indicators,
    misc_units,
    si_units,
    time_units,
)
from .represent.model import Model

__all__ = [
    "Currency",
    "Economic",
    "Emission",
    "Environ",
    "Land",
    "Linkage",
    "Location",
    "Material",
    "Model",
    "Periods",
    "Player",
    "Process",
    "Resource",
    "Social",
    "Storage",
    "TemporalScales",
    "Transport",
    "Unit",
    "currencies",
    "env_indicators",
    "misc_units",
    "si_units",
    "time_units",
]
__version__ = "2.0.1"



"""Storage - Stashes Resource to Withdraw Later"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._component import _Component
from ...modeling.constraints.calculate import Calculate
from ...modeling.parameters.conversion import Conversion
from ..commodity.stored import Stored
from .process import Process

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ...modeling.variables.sample import Sample
    from ..commodity.resource import Resource
    from ..spatial.location import Location


@dataclass
class Storage(_Component):
    """Storage

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: An optional citation or description for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional
    :param store: The resource to be stored. Defaults to None.
    :type store: Resource, optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar stored: The resource being stored. Defaults to None.
    :vartype stored: Stored, optional
    :ivar conv: Operational conversion associated with the storage. Defaults to None.
    :vartype conv: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar charge: The charge process associated with the storage.
    :vartype charge: Process
    :ivar discharge: The discharge process associated with the storage.
    :vartype discharge: Process
    :ivar locations: List of locations where the storage is located. Defaults to [].
    :vartype locations: list[Location]
    """

    store: Resource = None

    def __post_init__(self):
        _Component.__post_init__(self)
        self.stored: Stored = None
        self._conv = False
        self.conv = None
        self.charge = Process()
        self.charge.ofstorage = self
        self.discharge = Process()
        self.discharge.ofstorage = self
        self.locations: list[Location] = []

    def __setattr__(self, name, value):
        if name == "model" and value:
            setattr(value, f"{self.name}.charge", self.charge)
            setattr(value, f"{self.name}.discharge", self.discharge)

        super().__setattr__(name, value)

    def locate(self, *locations: Location):
        """Locate the storage"""
        # update the locations at which the storage exists

        # get location, time tuples where operation is defined
        for loc in locations:

            if self not in self.model.invcapacity.bound_spaces:
                self.model.invcapacity.bound_spaces[self.stored] = {"ub": [], "lb": []}

            if loc not in self.model.invcapacity.bound_spaces[self.stored]["ub"]:
                # check if the storage capacity has been bound at that location
                print(
                    f"--- Assuming  {self.stored} inventory capacity is unbounded in ({loc}, {self.horizon})",
                )
                # this is not a check, this generates a constraint
                _ = self.capacity(loc, self.horizon) == True

            if self.stored not in self.model.inventory.bound_spaces:
                _ = self.model.inventory(self.stored) == True

            if loc not in self.model.inventory.bound_spaces[self.stored]["ub"]:
                # check if the storage inventory has been bound at that location
                print(
                    f"--- Assuming inventory of {self.stored} is bound by inventory capacity in ({loc}, {self.horizon})",
                )
                try:
                    times = list(
                        [
                            t
                            for t in self.model.grb[self.stored.inv_of][loc]
                            if self.model.grb[self.stored.inv_of][loc][t]
                        ],
                    )
                except KeyError:
                    times = []
                # write the conversion balance at
                # densest temporal scale in that space
                if times:
                    time = min(times)
                else:
                    time = self.horizon
                # if not just write opr_{pro, loc, horizon} <= capacity_{pro, loc, horizon}
                print(
                    f"--- Assuming inventory of {self.stored} is bound by inventory capacity in ({loc}, {time})",
                )

                _ = self.inventory(loc, time) <= 1

        # locate the charge and discharge processes
        self.charge.locate(*locations)
        self.discharge.locate(*locations)

        # self.writecons_conversion(loc_times)

    @property
    def capacity(self) -> Sample:
        """Reports invcapacity as capacity"""
        return self.stored.invcapacity

    @property
    def setup(self) -> Sample:
        """Reports invsetup as setup"""
        return self.stored.invsetup

    @property
    def dismantle(self) -> Sample:
        """Reports invdismantle as dismantle"""
        return self.stored.invdismantle

    @property
    def inventory(self) -> Sample:
        """Inventory of the stored resource"""
        return self.stored.inventory

    @property
    def capex(self) -> Calculate:
        """Capital Expenditure"""
        return self.capacity[self.model.default_currency().spend]

    @property
    def opex(self) -> Calculate:
        """Operational Expenditure"""
        return self.charge.operate[self.model.default_currency().spend]

    @property
    def base(self) -> Resource:
        """Base resource"""
        return self.discharge.conv.base

    @property
    def conversion(self) -> dict[Resource, int | float]:
        """Conversion of commodities"""
        return self.discharge.conv.conversion

    @property
    def storage_cost(self) -> Calculate:
        """Cost of storing the resource"""
        return self.inventory[self.model.default_currency().spend]

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        # This overwrites the Component cons property
        # this gets the actual constraint objects from the program
        # based on the pname (attribute name) in the program
        return (
            [getattr(self.program, c) for c in self.constraints]
            + self.charge.cons
            + self.discharge.cons
            + self.stored.cons
        )

    @property
    def fab(self) -> Conversion:
        """Fabrication conversion of commodities"""
        return self.charge.fab

    def __call__(self, resource: Stored | Conversion):
        """Conversion is called with a Resource to be converted"""
        if not self._conv:
            # create storage resource
            stored = Stored()
            resource.in_inv.append(stored)

            setattr(self.model, f"{resource}.{self}", stored)

            # ---------- set discharge conversion
            self.discharge.conv = Conversion(operation=self.discharge, resource=stored)
            _ = self.discharge.conv(resource) == -1.0 * stored

            # ---------- set charge conversion
            self.charge.conv = Conversion(operation=self.charge, resource=resource)

            _ = self.charge.conv(stored) == -1.0 * resource

            setattr(self.base, self.name, stored)

            self.stored, self.stored.inv_of = stored, self.base
            self._conv = True

        return self.discharge.conv(resource)



"""Conversion"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Self

from ..._core._name import _Name
from ...components.temporal.lag import Lag
from ...components.temporal.modes import Modes

if TYPE_CHECKING:
    from gana.block.program import Prg

    from ...components.commodity.resource import Resource
    from ...components.operation._operation import _Operation
    from ...components.temporal.periods import Periods
    from ...represent.model import Model
    from ..variables.sample import Sample


@dataclass
class Conversion(_Name):
    """Processes convert one Resource to another Resource

    Conversion provides the conversion of resources

    Attributes:
        label (str): Label of the component, used for plotting. Defaults to None.
        process (Process): Process that converts Resources. Defaults to None.
        storage (Storage): Storage that stores Resources. Defaults to None.
        resource (Resource): Resource that is balanced. Defaults to None.
        name (str): generated as η(process).
        operation (Process | Storage): Process or Storage that serves as primary spatial index.
        base (Resource): Basis Resource, usually has a value 1 in the conversion matrix, set using __call__. Defaults to None.
        conversion (dict[Resource : int | float]): {Resources: conversion}. Defaults to {}, set using __eq__.
        lag (Lag): Temporal lag (processing time) for conversion, set using __getitem__. Defaults to None.

    Raises:
        TypeError: If __getitem__ input is not of type Lag.

    Note:
        - name and operation are generated post init
        - base (__call__), conversion (__eq__), lag (__getitem__) are defined as the program is built
        - Storage contains two processes (charge and discharge), hence is provided separately

    """

    resource: Resource = None
    operation: _Operation = None
    bind: Sample = None

    def __post_init__(self):
        self.name = f"η({self.operation})"
        self.base: Resource = None
        self.conversion: dict[Resource, int | float | list[int | float]] = {}
        self.lag: Lag = None
        self.periods: Periods = None

        # if piece wise linear conversion is provided
        self.pwl: bool = False

        # this is holds a mode for the conversion to be appended to
        self._mode: int | str = None

        # modes if PWL conversion is defined
        # or if multiple modes are defined
        self._modes: Modes = None

        # if the keys are converted into Modes
        self.modes_set: bool = False

    @property
    def modes(self) -> Modes:
        """Modes of the operation"""
        if self._modes is None:
            n_modes = len(self.conversion)
            modes_name = f"bin{len(self.model.modes)}"

            setattr(self.model, modes_name, Modes(n_modes=n_modes, bind=self.bind))

            self._modes = self.model.modes[-1]

        return self._modes

    @cached_property
    def model(self) -> Model:
        """energia Model"""

        if self.pwl:
            _conversion = self.conversion[next(iter(self.conversion))]
        else:
            _conversion = self.conversion

        return next((i.model for i in _conversion), None)

    @cached_property
    def program(self) -> Prg:
        """gana Program"""
        return self.operation.program

    def balancer(self):
        """Checks if there is a list in the conversion
        If yes, tries to make everything consistent
        """

        def _balancer(conversion: dict):

            # check if lists are provided
            check_list = dict.fromkeys(conversion.keys(), False)
            # check lengths of the list, for parameter the length is 1
            check_len = dict.fromkeys(conversion.keys(), 1)

            for res, par in conversion.items():
                if isinstance(par, list):
                    check_list[res] = True
                    check_len[res] = len(par)

            # check if all the list lens are the same
            lengths = {i for i in check_len.values() if i > 1}

            if len(lengths) > 1:
                # if there are different lengths, raise an error
                raise ValueError(
                    f"Conversion: {self.name} has inconsistent list lengths: {lengths}",
                )

            if any(check_list.values()):
                length = next(iter(lengths))
                # if any of the values are a list
                #
                for res, par in conversion.items():
                    if isinstance(par, (float, int)):
                        conversion[res] = [par] * length
            return conversion

        if self.pwl:
            for mode, conv in self.conversion.items():
                self.conversion[mode] = _balancer(conv)

            if isinstance(next(iter(self.conversion)), Modes):
                self.modes_set = True

        else:
            self.conversion = _balancer(self.conversion)

    def __getitem__(self, mode: int | str) -> Self:
        """Used to define mode based conversions"""
        self._mode = mode
        return self

    def __call__(self, basis: Resource | Conversion, lag: Lag = None) -> Self:
        # sets the basis
        if isinstance(basis, Conversion):
            # if a Conversion is provided (parameter*Resource)
            # In this case the associated conversion is not 1
            # especially useful if Process is scaled to consumption of a resource
            # i.e. basis = -1*Resource
            self.conversion = {**self.conversion, **basis.conversion}
            self.base = next(iter(self.conversion))

        else:
            # if a Resource is provided (Resource)
            # implies that the conversion is 1
            # i.e the Process is scaled to one unit of this Resource produced
            self.base = basis
            self.conversion = {basis: 1.0, **self.conversion}

        if lag:
            self.lag = lag

        return self

    def __eq__(self, other: Conversion | int | float | dict[int | float, Conversion]):
        # cons = []

        if isinstance(other, (int, float)):
            # this is used for inventory conversion
            # when not other resource besides the one being inventoried is involved

            self.conversion = {**self.conversion, self.resource: -1.0 / float(other)}

        elif isinstance(other, dict):

            key = next(iter(other.keys()))

            if isinstance(key, Modes):
                # conversion modes can collate
                # for example resource conversion modes and material conversion modes
                self.modes_set = True
                self._modes = key.parent

            # this is when there is a proper resource conversion
            # -20*res1 = 10*res2 for example
            self.conversion = {
                k: {**self.conversion, **v.conversion} for k, v in other.items()
            }
            self.pwl = True

        # this would be a Conversion or Resource
        elif self._mode is not None:
            self.conversion[self._mode] = other.conversion
            if not self.pwl:
                self.pwl = True
            self._mode = None
        else:

            self.conversion: dict[Resource, int | float] = {
                **self.conversion,
                **other.conversion,
            }

        self.model.convmatrix[self.operation] = self.conversion

    # these update the conversion of the resource (self.conversion)
    def __add__(self, other: Conversion) -> Self:
        if isinstance(other, Conversion):
            self.conversion = {**self.conversion, **other.conversion}
            return self
        self.conversion = {**self.conversion, other: 1}
        return self

    def __sub__(self, other: Conversion) -> Self:
        if isinstance(other, Conversion):
            self.conversion = {
                **self.conversion,
                **{res: -1 * par for res, par in other.conversion.items()},
            }
            return self
        self.conversion = {**self.conversion, other: -1}
        return self

    def __mul__(self, times: int | float | list) -> Self:
        if isinstance(times, list):
            self.conversion = {
                res: [par * i for i in times] for res, par in self.conversion.items()
            }
        else:
            self.conversion = {res: par * times for res, par in self.conversion.items()}
        return self

    def __rmul__(self, times) -> Self:
        return self * times

    def __truediv__(self, periods: Periods) -> Self:
        self.periods = periods
        return self



"""Inherited _Component class"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from gana import I

from ._x import _X

if TYPE_CHECKING:
    from ..components.measure.unit import Unit
    from ..components.spatial.location import Location
    from ..components.temporal.periods import Periods
    from ..dimensions.problem import Problem
    from ..dimensions.space import Space
    from ..dimensions.time import Time


@dataclass
class _Component(_X):
    """
    A component with a mathematical program (hence index in it) and
    input parameters (modeling aspects).

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: Label for the component. Defaults to None.
    :type label: str, optional
    :param captions: Citation for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]

    :note:
        - `name`, `model` set when component assigned to Model attribute.
        - `constraints` and `domains` are populated as the program is built.
    """

    # every component has a basis unit
    basis: Optional[Unit] = None

    def __post_init__(self):

        # what differentiates a component from an index is that it has aspects
        # that we can control to adjust their states of existence
        _X.__post_init__(self)

    @property
    def network(self) -> Location:
        """Circumscribing Loc (Spatial Scale)"""
        return self.model.network

    @property
    def horizon(self) -> Periods:
        """Circumscribing Periods (Temporal Scale)"""
        return self.model.horizon

    @cached_property
    def time(self) -> Time:
        """Time"""
        return self.model.time

    @cached_property
    def space(self) -> Space:
        """Space"""
        return self.model.space

    @cached_property
    def problem(self) -> Problem:
        """Feasible Region"""
        # the decision tree gives the component access to
        # the aspects of all other components
        return self.model.problem

    def __getattr__(self, name):

        if self.model:
            # no need to run a hasattr check
            # let it raise an attribute error if not found
            aspect = getattr(self.model, name)
            if callable(aspect):
                return aspect(self)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'",
        )

    @cached_property
    def I(self):
        """gana index set (I)"""
        _index = I(self.name, tag=self.label or "")
        setattr(self.program, self.name, _index)
        return _index



"""Operational Mode attached to a Parameter"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Self

from gana import I

from ..._core._x import _X

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ...modeling.variables.sample import Sample


@dataclass
class Modes(_X):
    """
    Represents a discrete choice to be taken within a spatiotemporal disposition.
    Modes can split usage. A Mode of Operation can be used for Conversion, Use, etc.

    :param n_modes: Number of modes. Defaults to 1.
    :type n_modes: int
    :param bind: The aspect and component which is being 'moded'. Defaults to None.
    :type bind: Bind | None
    :param parent: Parent mode, if any. Defaults to None.
    :type parent: Self | None
    :param pos: Position in the parent mode set. Defaults to None.
    :type pos: int | None

    :ivar name: The name of the mode, usually a number.
    :vartype name: str | float | int
    """

    n_modes: int = 1
    bind: Sample = None
    parent: Self = None
    pos: int = None

    def __post_init__(self):
        _X.__post_init__(self)

        if self.pos is not None:
            self.name = f"{self.parent}[{self.pos}]"

        if self.parent:
            self.model = self.parent.model

        # have child modes been made
        self._birthed = False
        # where child modes are stored
        self._ = []

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        # overwrite X.cons property
        # this gets the actual constraint objects from the program
        # based on the pname (attribute name) in the program
        if self.parent:
            return [getattr(self.program, c) for c in self.constraints]
        return [getattr(self.program, c) for c in self.constraints] + sum(
            [m.cons for m in self],
            [],
        )

    @cached_property
    def I(self) -> I:
        """Index set of modes"""

        if self.parent:
            _ = self.parent.I  # makes sure the parent is indexed
            # do not set a new index set, get from parent
            return getattr(self.parent.program, self.parent.name)[self.pos]

        _index = I(size=self.n_modes, tag=f"Modes of {self.bind.aspect}")

        setattr(
            self.program,
            self.name,
            _index,
        )

        return _index

    def __eq__(self, other: Modes) -> bool:
        """Equality operator"""
        if other.name == self.name:
            return True
        return False

    def __len__(self) -> int:
        """Length of the modes"""
        return self.n_modes

    def __getitem__(self, key: int) -> _X:
        """Get a mode by index"""
        if not self._birthed:
            self._ = [
                Modes(bind=self.bind, parent=self, pos=i) for i in range(self.n_modes)
            ]
            self._birthed = True

        return self._[key]

    def __iter__(self):
        """Iterate over modes"""
        for i in range(self.n_modes):
            yield self[i]



"""Variable"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Optional, Self, Type

import matplotlib.pyplot as plt
from matplotlib import rc

from ...components.commodity._commodity import _Commodity
from ...components.game.couple import Couple
from ...components.game.player import Player
from ...components.impact.indicator import Indicator
from ...components.operation.process import Process
from ...components.operation.storage import Storage
from ...components.operation.transport import Transport
from ...components.spatial.linkage import Linkage
from ...components.spatial.location import Location
from ...components.temporal.lag import Lag
from ...components.temporal.modes import Modes
from ...components.temporal.periods import Periods
from ...dimensions.space import Space
from ...dimensions.time import Time
from ..indices.domain import Domain
from .sample import Sample

if TYPE_CHECKING:
    from gana import I as Idx
    from gana import Prg
    from gana import V as Var
    from gana.sets.constraint import C

    from ..._core._component import _Component
    from ..._core._x import _X
    from ...dimensions.problem import Problem
    from ...modeling.indices.domain import Domain
    from ...represent.model import Model


@dataclass
class Aspect:
    r"""
    A particular facet of the system under consideration. A sample of an aspect at a
    specific disposition is represented by a variable (:math:`\overset{*}{v} \in \overset{*}{\mathcal{V}}`).
    The range of values which the :math:`\overset{*}{v}` is bounded such that:

    .. math::

        \overset{*}{v} \in [\underline{\theta}, \overline{\theta}]

    Aspects can be decision-variables (:math:`\dot{v}`) that prescribe a control action
    or a set point for the state, or derived (calculated) variables (:math:`\hat{v}`) that
    represent the state of the system. States can be operation capacity, production levels,
    purchase levels, emissions, consumption levels, etc.


    :param primary_types: associated components type(s). Defaults to None.
    :type primary_types: Type[_Component] | tuple[Type[_Component], ...]
    :param nn: If True, the decision is a non-negative decision. Defaults to True.
    :type nn: bool
    :param ispos: If True, the decision is positive (non-negative). Defaults to True.
    :type ispos: bool
    :param neg: Negative form or representation of the decision, if any. Defaults to "".
    :type neg: str
    :param latex: LaTeX string. Defaults to "".
    :type latex: str
    :param bound: if the aspect is bounded by another. Defaults to "".
    :type bound: str
    :param label: Label for the decision. Defaults to "".
    :type label: str

    :ivar model: Model to which the Aspect belongs.
    :vartype model: Model
    :ivar name: Name of the Aspect.
    :vartype name: str
    :ivar indices: List of indices (Location, Periods) associated with the Aspect.
    :vartype indices: list[Location | Linkage, Periods]
    :ivar bound_spaces: Spaces where the Aspect has been already bound.
    :vartype bound_spaces: dict[_Commodity | Process | Storage | Transport, list[Location | Linkage]]
    :ivar domains: List of domains associated with the Aspect.
    :vartype domains: list[Domain]

    raises ValueError:
        - If `primary_type` is not defined.
    """

    primary_type: Type[_Component] | tuple[Type[_Component], ...]
    nn: bool = True
    ispos: bool = True
    neg: str = ""
    latex: str = ""
    bound: str = ""
    label: Optional[str] = None

    def __post_init__(self):
        # will be set when added to model
        self.name: str = ""

        # name of the decision
        self.model: Optional[Model] = None

        if self.label:
            if self.nn:
                self.label += " [+]"
            else:
                self.label += " [-]"
        self.indices: list[Location | Linkage | Periods] = []

        # # if a decision is bounded by another decision
        # self.bound: Self = None

        # spaces where the aspect has been already bound
        self.bound_spaces: dict[
            _Commodity | Process | Storage | Transport,
            list[Location | Linkage],
        ] = {}

        # upper/lower/exact bounds are set on these locations/periods

        self.ubs: dict[Location | Linkage, Periods] = {}
        self.lbs: dict[Location | Linkage, Periods] = {}
        self.eqs: dict[Location | Linkage, Periods] = {}

        # Domains of the decision
        self.domains: list[Domain] = []

        # a dictionary of domains and their maps from higher order domains

        self._maps: bool = False

        self._maps_report: bool = False

        # reporting variable
        self.reporting: Optional[Var] = None

        self.constraints: list[str] = []

        # this keeps track of whether GRB has already been added
        self.grb: dict[tuple[Idx, ...], bool] = {}

    @property
    def maps(self) -> dict[Domain, dict[str, list[Domain]]]:
        """Maps of the decision"""
        if not self._maps:
            self.model.maps[self] = {}
            self._maps = True
        return self.model.maps[self]

    @property
    def maps_report(self) -> dict[Domain, dict[str, list[Domain]]]:
        """Maps of the decision"""
        if not self._maps_report:
            self.model.maps_report[self] = {}
            self._maps_report = True
        return self.model.maps_report[self]

    @cached_property
    def isneg(self) -> bool:
        """Does this remove from the domain?"""
        return not self.ispos

    @cached_property
    def sign(self) -> float:
        """Gives the multiplier in balances"""
        if self.ispos:
            return 1.0
        else:
            return -1.0

    @cached_property
    def space(self) -> Space:
        """Space"""
        return self.model.space

    @cached_property
    def program(self) -> Prg:
        """Mathematical Program"""
        return self.model.program

    @cached_property
    def problem(self) -> Problem:
        """Tree"""
        return self.model.problem

    @cached_property
    def time(self) -> Time:
        """Time"""
        return self.model.time

    @property
    def I(self) -> Idx:
        """gana index set (I)"""
        return getattr(self.program, self.name)

    @property
    def V(self) -> Var:
        """Variable"""
        return getattr(self.program, self.name)

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        return [getattr(self.program, c) for c in self.constraints]

    @property
    def network(self) -> Location:
        """Circumscribing Location (Spatial Scale)"""
        return self.model.network

    @property
    def horizon(self) -> Periods:
        """Circumscribing Periods (Temporal Scale)"""
        return self.model.horizon

    @property
    def dispositions(self) -> dict[
        Self,
        dict[
            _Commodity | Process | Storage | Transport,
            dict[
                Periods | Location | Linkage,
                dict[Location | Periods | Linkage, bool],
            ],
        ],
    ]:
        """Dispositions dict"""
        return self.model.dispositions[self]

    def aliases(self, *names: str):
        """Create aliases for the decision

        Args:
            *names (str): Names of the aliases
        """
        self.model.aliases(*names, to=self.name)

    def update(self, domain: Domain, reporting: bool = False):
        """Each inherited object has their own"""

    def show(self, descriptive=False):
        """Pretty print the component"""
        for c in self.cons:
            c.show(descriptive)

    def sol(
        self,
        n_sol: int = 0,
        aslist: bool = False,
        compare: bool = False,
    ) -> list[float] | None:
        """Solution
        Args:
            aslist (bool, optional): Returns values taken as list. Defaults to False.
        """
        var: Var = getattr(self.program, self.name)
        return var.sol(n_sol, aslist, compare=compare)

    def gettime(self, *index) -> list[Periods]:
        """Finds the sparsest time scale in the domains"""
        ds = [i for i in self.indices if all([x in i for x in index])]
        t = [t for t in ds if isinstance(t, Periods)]
        return t

    def draw(
        self,
        x: _X,
        y: tuple[_X] | _X,
        z: tuple[_X] | _X = None,
        font_size: float = 16,
        fig_size: tuple[float, float] = (12, 6),
        linewidth: float = 0.7,
        color: str = "blue",
        grid_alpha: float = 0.3,
        usetex: bool = True,
    ):
        """Plot the decision"""
        if not isinstance(y, tuple):
            y = (y,)
        if not z:
            index = [i.I for i in y + (x,)]
        elif not isinstance(z, tuple):
            z = (z,)

        if usetex:
            rc(
                "font",
                **{"family": "serif", "serif": ["Computer Modern"], "size": font_size},
            )
            rc("text", usetex=usetex)
        else:
            rc("font", **{"size": font_size})

        fig, ax = plt.subplots(figsize=fig_size)

        if not z:
            ax.plot(
                [i.name for i in x.I._],
                self.V(*index).sol(True),
                linewidth=linewidth,
                color=color,
            )
        else:
            for z_ in z:
                index = [i.I for i in (z_,) + y + (x,)]
                ax.plot(
                    [i.name for i in x.I._],
                    self.V(*index).sol(True),
                    linewidth=linewidth,
                    label=z_.label if z_.label else z_.name,
                )

        label_ = [i.label if i.label else i.name for i in y]
        label_ = str(label_).replace("[", "").replace("]", "").replace("'", "")
        plt.title(f"{self.label} - {label_}")
        plt.ylabel("Values")
        plt.xlabel(f"{x.label or x.name}")
        plt.grid(alpha=grid_alpha)
        if z:
            plt.legend()
        plt.rcdefaults()

    def __neg__(self):
        """Negative Consequence"""

        dscn = type(self)(
            nn=False,
            primary_type=self.primary_type,
        )
        dscn.neg, self.neg = self, dscn
        dscn.ispos = not self.ispos
        return dscn

    def __len__(self):
        return len(self.domains)

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, Aspect):
            return self.name == other.name

    def __call__(self, *index: _X, domain: Domain = None):

        if not domain:

            args = {
                "indicator": None,
                "commodity": None,
                "player": None,
                "process": None,
                "storage": None,
                "transport": None,
                "periods": None,
                "couple": None,
                "location": None,
                "linkage": None,
                "lag": None,
                "modes": None,
            }

            type_map = {
                Periods: ("periods", "timed", False),
                Lag: ("lag", "timed", False),
                Location: ("location", "spaced", False),
                Linkage: ("linkage", "spaced", False),
                Process: ("process", None, True),
                Storage: ("storage", None, True),
                Transport: ("transport", None, True),
                Player: ("player", None, False),
                Couple: ("couple", None, False),
                Indicator: ("indicator", None, False),
                Modes: ("modes", None, False),
                _Commodity: ("commodity", None, True),
            }
            # (
            #     indicator,
            #     commodity,
            #     player,
            #     process,
            #     storage,
            #     transport,
            #     period,
            #     couple,
            #     location,
            #     linkage,
            #     lag,
            #     modes,
            # ) = (None for _ in range(12))

            binds: list[Sample] = []
            timed, spaced = False, False

            for comp in index:
                if isinstance(comp, Sample):
                    binds.append(comp)
                    for b in binds:
                        if b.domain.binds:
                            binds.extend(b.domain.binds)
                    binds = list(set(binds))
                    continue

                for typ, (attr, flag, require_primary) in type_map.items():
                    if isinstance(comp, typ):
                        if require_primary and (
                            not self.primary_type
                            or not isinstance(comp, self.primary_type)
                        ):
                            raise ValueError(
                                f"For component {self} of type {type(self)}: "
                                f"{comp} of type {type(comp)} not recognized as an index",
                            )
                        args[attr] = comp
                        if flag == "timed":
                            timed = True
                        elif flag == "spaced":
                            spaced = True
                        break
                else:
                    raise ValueError(
                        f"For component {self} of type {type(self)}: "
                        f"{comp} of type {type(comp)} not recognized as an index",
                    )

            args = {k: v for k, v in args.items() if v is not None}

            if binds:
                args["binds"] = binds

            domain = Domain(**args)

        else:
            timed = spaced = True

        return Sample(aspect=self, domain=domain, timed=timed, spaced=spaced)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __init_subclass__(cls):
        cls.__repr__ = Aspect.__repr__
        cls.__hash__ = Aspect.__hash__



"""Function for scaling input data"""

from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler, StandardScaler


def scaling(data: DataFrame, how: str) -> DataFrame:
    """Scales input DataFrame

    Args:
        data (DataFrame): input data
        how (str): 'max', 'minmax', 'standard'

    Returns:
        DataFrame: scaled DataFrame

    Raises:
        ValueError: if data is not a DataFrame

    """
    if not isinstance(data, DataFrame):
        raise ValueError("please provide DataFrame")

    if how == "max":
        data = data / data.max()

    else:
        if how == "standard":
            scaler = StandardScaler()
        if how == "minmax":
            scaler = MinMaxScaler()
        data = DataFrame(scaler.fit_transform(data))

    return data



"""Temporal Lag"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING, Self

from ..._core._name import _Name

if TYPE_CHECKING:
    from gana import I as Idx

    from ...modeling.indices.domain import Domain
    from .periods import Periods


@dataclass
class Lag(_Name):
    """
    A number of temporal Periods.

    :ivar label: Label of the component, used for plotting. Defaults to None.
    :vartype label: str
    :ivar of: Periods to lag. Defaults to None.
    :vartype of: Periods
    :ivar periods: Number of periods to lag. Defaults to 1.
    :vartype periods: int | float
    :ivar name: Name of the component, generated based on the number of periods.
    :vartype name: str
    :ivar domains: List of Domains the lag features belong to. Defaults to [].
    :vartype domains: list[Domain]

    .. note::
    - Name is generated post init.
    - Domains are set as the program is built.
    """

    of: Periods = None
    periods: int | float = 1

    def __post_init__(self):
        self.name = f"-{self.periods}{self.of}"
        self.domains: list[Domain] = []
        self.constraints: list[str] = []

    @cached_property
    def I(self) -> Idx:
        """Index set of periods - periods
        Basically,
        if t is set of hours and lag is 5
        this returns the set of t - 5
        i.e. an offset of 5 hours
        """
        return self.of.I - self.periods

    @property
    def horizon(self) -> Periods:
        """Horizon of the lag"""
        return self.of.horizon

    def __eq__(self, other: Self | Periods) -> bool:
        """Check if two lags are equal"""
        if isinstance(other, Lag):
            return is_(self.of, other.of) and self.periods == other.periods
        return is_(self.of, other)



"""Node"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._name import _Name

if TYPE_CHECKING:
    from ...represent._ations.graph import Graph


@dataclass
class Node(_Name):
    """Node of a Graph

    Attributes:
        label (str): Label of the component, used for plotting. Defaults to None.
        graph (Graph): Graph to which the node belongs. Defaults to None.
        name (str): name. Defaults to ''.

    Note:
        - name and Graph are set when made a Graph attribute.
    """

    def __post_init__(self):
        self.graph: Graph = None
        _Name.__post_init__(self)



"""Decision Space of the Model"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..components.commodity.currency import Currency
from ..components.commodity.emission import Emission
from ..components.commodity.land import Land
from ..components.commodity.material import Material
from ..components.commodity.resource import Resource
from ..components.commodity.stored import Stored
from ..components.impact.categories import Economic, Environ, Social
from ..components.operation.process import Process
from ..components.operation.transport import Transport
from ..modeling.variables.control import Control
from ..modeling.variables.states import (
    EndoStream,
    ExoStream,
    IndStream,
    SetPoint,
    Size,
    State,
)

if TYPE_CHECKING:
    from ..represent.model import Model


def capacity_sizing(model: Model):
    """Sets capacity sizing decisions"""
    model.Recipe(
        "capacity",
        Size,
        primary_type=(Process, Transport),
        label="Capacitate Operation",
        latex=r"{cap}",
        add_kind=Control,
        add="setup",
        add_latex=r"{cap}^{+}",
        sub="dismantle",
        sub_latex=r"{cap}^{-}",
    )


def operating(model: Model):
    """Operating/Scheduling decisions"""

    model.Recipe(
        "operate",
        SetPoint,
        primary_type=(Process, Transport),
        label="Capacity Utilization",
        latex=r"{opr}",
        add_kind=Control,
        add="rampup",
        add_latex=r"{opr}^{+}",
        sub="rampdown",
        sub_latex=r"{opr}^{-}",
        bound="capacity",
    )

    model.Recipe(
        "produce",
        EndoStream,
        primary_type=Resource,
        latex="prod",
        label="Stream produced in Operation",
        neg="expend",
        neg_latex=r"{expd}",
        neg_label="Stream expended in Operation",
    )

    model.Recipe(
        "ship_in",
        EndoStream,
        primary_type=Resource,
        label="Resource Imported into Location",
        latex=r"{impt}",
        neg="ship_out",
        neg_label="Resource Exported out of Location",
    )


def inventory_sizing(model: Model):
    """Inventory management decisions"""
    model.Recipe(
        "invcapacity",
        Size,
        primary_type=Resource,
        label="Inventory Capacity",
        latex=r"{icap}",
        add_kind=Control,
        add="invdismantle",
        add_latex=r"{icap}^{-}",
        sub="invsetup",
        sub_latex=r"{icap}^{+}",
    )

    model.Recipe(
        "inventory",
        EndoStream,
        primary_type=Stored,
        label="Stored Resource",
        latex=r"{inv}",
        ispos=False,
        bound="invcapacity",
    )


def usage(model: Model):
    """Resource usage decisions"""
    model.Recipe(
        "dispose",
        EndoStream,
        primary_type=(Resource, Land, Material),
        label="Dispose Resource",
        latex=r"{disp}",
        neg="use",
        neg_label="Use Resource",
    )


def free_movement(model: Model):
    """Free movement entails consumption or release of resources

    This is an interaction with the system exterior

    .. note::
        Use buy and sell if interacting with a Player

    """
    model.Recipe(
        "consume",
        ExoStream,
        primary_type=Resource,
        label="Free Resource Stream",
        latex=r"{cons}",
        add_kind=Control,
        neg="release",
        neg_latex=r"{rlse}",
    )


def trade(model: Model):
    """Trade decisions"""
    model.Recipe(
        "buy",
        ExoStream,
        primary_type=Resource,
        label="Buy Resource",
        neg="sell",
    )


def economic(model: Model):
    """Economic decisions"""
    model.Recipe(
        "earn",
        IndStream,
        primary_type=(Currency, Economic),
        neg="spend",
    )


def environmental(model: Model):
    """Environmental impact decisions"""
    model.Recipe(
        "emit",
        IndStream,
        primary_type=(Environ, Emission),
        neg="abate",
    )


def social(model: Model):
    """Social impact decisions"""
    model.Recipe(
        "benefit",
        IndStream,
        primary_type=(Social, Currency),
        neg="detriment",
    )



"""Unit"""

from operator import is_
from typing import Optional, Self

from ..._core._name import _Name


class Unit(_Name):
    """Unit of measure for a quantity provided as input to a component

    :param label: Label of the component, used for plotting. Defaults to None.
    :type label: str, optional
    :param basis: Basis of the unit, if defined based on another Unit. Defaults to None.
    :type basis: Unit, optional
    :param times: How many times the basis is contained in the unit. Defaults to None.
    :type times: int | float, optional

    :ivar name: Name of the object. Defaults to ''.
    :vartype name: str
    """

    def __init__(
        self,
        basis: Optional[Self] = None,
        times: Optional[int | float] = None,
        label: Optional[str] = None,
    ):
        """
        Args:
            basis (Unit): Basis of the unit, if defined based on another Unit. Defaults to None.
            times (int | float): How many times the basis is contained in the unit. Defaults to None.
            label (str): Label of the unit. Defaults to None.
            name (str): Name of the unit. Defaults to ''.

        Note:
            - name is set when the component is made a Model attribute.
        """
        # A basis can itself be measured using another basis
        self.basis = basis
        # How many times that basis is self?
        self.times = times

        _Name.__init__(self, label)

    def howmany(self, basis: Self):
        """How many times is this basis contained in the other basis

        Args:
            basis (Unit): Basis to compare with.
        """
        # if other Unit is basis of self
        # return the times of self
        if is_(basis, self.basis):
            return self.times

        # if self is the basis of the other Unit
        # return the times of the other Unit
        elif is_(self, basis.basis):
            return basis.times

        # if both have a common basis
        # return the times of self divided by the times of the other Unit
        elif is_(self.basis, basis.basis):
            return self.times / basis.times

        # if no common basis is found
        # raise an error
        raise ValueError(
            f"{self} and {basis} do not have a common basis for comparison",
        )

    def __truediv__(self, other: float):
        # can only be divided by a number
        # if dividing by another Unit
        # return the times of self divided by the times of the other Unit
        if isinstance(other, Unit):
            return self.howmany(other)

        # if dividing by a number
        # create a new Unit
        elif isinstance(other, (int, float)):
            # is self has a basis Unit
            # define new Unit based on the basis of self
            if self.basis:
                b = Unit(
                    self.basis,
                    1 / (other * self.times),
                    f"{self.label}/{other*self.times}",
                )
            # if self has no basis Unit
            # define new Unit using self as basis
            else:
                b = Unit(self, 1 / other, f"{self.label}/{other}")
                self.basis = b
                self.times = other
            return b

    def __mul__(self, other: float):
        # can only be multiplied by a number
        if not isinstance(other, (int, float)):
            raise TypeError(f"Cannot multiply Unit by {type(other)}")

        # if multiplying by a number
        # create a new Unit

        # if self has a basis Unit
        # define new Unit based on the basis of self
        if self.basis:
            b = Unit(self.basis, other * self.times, f"{self.label}.{other*self.times}")
        # if self has no basis Unit
        # define new Unit using self as basis
        else:
            b = Unit(self, other, f"{self.label}.{other}")
            self.basis = b
            self.times = 1 / other
        return b

    def __rmul__(self, other: float):
        return self * other



"""Program"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from gana.block.program import Prg
from gana.sets.index import I

if TYPE_CHECKING:
    from ..model import Model


@dataclass
class Program(Prg):
    """The mathematical program of the model
    Attributes:
        players (I): Players index set
        periods (I): Periods index set
        locations (I): Locations index set
        linkages (I): Linkages index set
        sources (I): Sources index set
        sinks (I): Sinks index set
        spaces (I): Spaces index set
        currencies (I): Currencies index set
        envs (I): Environments index set
        socs (I): Societies index set
        ecos (I): Ecosystems index set
        resources (I): Resources index set
        lands (I): Lands index set
        materials (I): Materials index set
        processes (I): Processes index set
        storages (I): Storages index set
        transits (I): Transits index set
        states (I): States index set
        controls (I): Decisions index set
        streams (I): Streams index set
        impacts (I): Consequences index set

    Note:
        - all the index sets are generated post initialization
    """

    model: Model = None

    def __post_init__(self):
        Prg.__post_init__(self)

        # Component Index Sets
        self.name = f"Program({self.model})"

    def __getattr__(self, item):

        if item in self.model.dimension_map:
            index = I(mutable=True)
            setattr(self, item, index)
            return index
        raise AttributeError(
            f"{self} has no '{item}'",
        )



"""Time Periods"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING, Self

from gana import I as Idx

from ..._core._x import _X
from ...modeling.parameters.value import Value
from .lag import Lag

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ...dimensions.time import Time


@dataclass
class Periods(_X):
    """
    A discretization of Time.

    :param periods: Number of periods in Periods. Defaults to 1.
    :type periods: int | float
    :param of: The periods of which this is a multiple. Defaults to None.
    :type of: Periods | Lag, optional
    :param name: Name of the periods. Defaults to None.
    :type name: str, optional
    :param label: Label of the periods. Defaults to None.
    :type label: str, optional

    :ivar model: Model to which the Periods belongs.
    :vartype model: Model
    :ivar time: Time to which the Periods belongs.
    :vartype time: Time
    :ivar horizon: Horizon of the Time.
    :vartype horizon: Periods
    :ivar I: Index set of the Periods.
    :vartype I: I
    :ivar constraints: List of constraints associated with the Periods. Defaults to [].
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the Periods. Defaults to [].
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the Periods.
    :vartype aspects: dict[Aspect, list[Domain]]
    """

    periods: int | float = 1
    of: Self = None

    def __post_init__(self):

        _X.__post_init__(self)

        self._periods = self.periods

        self._of = self.of

        if self.of is not None and not self.of.isroot():
            self.periods = self.periods * self.of.periods
            self.of = self.of.of

        self._horizon: Self = None

        # can be overwritten by program
        self.name = f"{self._periods}{self._of}"

        if self.of is None:
            self.of = self

        # if this is a slice of another period
        self.slice: slice = None

        # if this is a single period in Periods
        self.pos: int = None

    def isroot(self):
        """Is used to define another period?"""
        if self.of is None:
            return True

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        return [getattr(self.program, c) for c in self.constraints]

    @property
    def time(self) -> Time:
        """Time to which the Periods belongs"""
        return self.model.time

    @property
    def horizon(self) -> Self:
        """Time Horizon"""
        return self.time.horizon

    @property
    def ishorizon(self) -> bool:
        """Is this the horizon of the model?"""
        return self == self.time.horizon

    @cached_property
    def I(self) -> Idx:
        """Index set of scale"""

        # given that temporal scale is an ordered set and not a self contained set
        # any time periods will be a fraction of the horizon
        if self.slice is not None:
            return self.of.I[self.slice]

        if self.pos is not None:
            return self.of.I[self.pos]

        _index = Idx(size=self.time.horizon.howmany(self), tag=self.label or "")
        setattr(self.program, self.name, _index)
        return _index

    def howmany(self, period: Self | Lag):
        """How many periods make this period"""

        if isinstance(period, Lag):
            # lags are usually made up of negative multiplications of periods
            return self.howmany(period.of) * period.periods

        if period == self:
            # if the same
            return 1

        if period.isroot():

            # if there are multiple root periods defined
            if is_(period, self.of):
                # if they are multiples of the same base period
                # check if they contain the same number of periods
                return self.periods
            raise ValueError(f"{period} is not a period of {self.name}")

        if is_(period.of, self.of):

            # if they are multiples of the same base period
            # return the ratio of the periods
            p = self.periods / period.periods
            if p.is_integer():
                return int(p)
            return p

        if is_(self, period.of):

            # if self is the base of the period
            # return the inverse of the of the periods
            return 1 / period.periods

        raise ValueError(f"{period} is not a period of {self.name}")

    def __mul__(self, times: int | float):

        if times < 0:
            # if multiplying by a negative number
            # return a lag
            if not self._of:
                return Lag(of=self, periods=-times)

            raise ValueError(f"{self} is not a period of anything, so cannot be lagged")

        if times < 1:
            # if it is a fraction
            period = Periods()
            if not self.of:
                # this is a root period
                period.of = self
                period.periods = times
            else:
                self.of = period
                self.periods = int(1 / times)
            return period

        return Periods(periods=int(times), of=self)

    def __truediv__(self, other: int | float):
        if isinstance(other, Periods):
            return self.howmany(other)

        return self * (1 / other)

        # return Periods(periods=other, of=self)

    def __rtruediv__(self, other: int | float):
        return Value(value=other, periods=self)

    def __rmul__(self, other: int | float):
        return self * other

    def __call__(self, times):
        return Periods(periods=times, of=self)

    def __neg__(self):
        return self * -1

    def __len__(self):
        return int(self.howmany(self.time.horizon))

    def __eq__(self, other: Self | Lag):
        if isinstance(other, Lag):
            return is_(self, other.of)

        return is_(self, other)

    def __ge__(self, other: Self):
        return self.time.horizon.howmany(self) <= self.time.horizon.howmany(other)

    def __gt__(self, other: Self):
        return self.time.horizon.howmany(self) < self.time.horizon.howmany(other)

    def __le__(self, other: Self):
        return self.time.horizon.howmany(self) >= self.time.horizon.howmany(other)

    def __lt__(self, other: Self):
        return self.time.horizon.howmany(self) > self.time.horizon.howmany(other)

    def __getitem__(self, key: int | slice):
        periods = Periods()
        periods.of = self
        periods.model = self.model

        if isinstance(key, slice):
            periods.slice = key
            start = key.start or 0
            stop = key.stop or len(self)
            periods.name = rf"{self}[{start}: {stop}]"
        else:
            periods.pos = key
            periods.name = rf"{self}[{key}]"

        return periods



"""dictionary utils"""

from collections import defaultdict


def get_depth(d: dict) -> int:
    """
    Finds the depth of a dictionary.

    Args:
        d (dict): The dictionary to measure.

    Returns:
        int: Depth of the dictionary.
    """
    if not isinstance(d, dict) or not d:
        # If the input is not a d or is an empty d, return 0
        return 0

    # Recursively find the maximum depth of nested dictionaries
    max_depth = max(get_depth(value) for value in d.values())

    # Return one more than the maximum depth found
    return 1 + max_depth


def flatten(d: dict, key: tuple = ()) -> dict:
    """
    Makes a flat dictionary from a nested dictionary.

    Args:
        d (dict): The dictionary to flatten.
        key (tuple, optional): Current key path. Defaults to ``()``.

    Returns:
        dict: Flattened dictionary with tuple keys.
    """
    items = []
    for key, val in d.items():
        keyupd = key + (key,)
        if isinstance(val, dict):
            items.extend(flatten(val, keyupd).items())
        else:
            items.append((keyupd, val))
    return dict(items)


def tupler(d: dict, path: tuple = ()) -> list[tuple[str]]:
    """
    Makes a list of tuples of keys in a nested dictionary.

    Parameters
    ----------
    d : :class:`dict`
        The dictionary to traverse.
    current_path : :class:`tuple`, optional
        Path taken to get to a value. Defaults to ``()``.

    Returns
    -------
    :class:`list` of :class:`tuple` of :class:`str`
        List of tuples of keys representing the paths to each value.
    """

    result = []

    for k, v in d.items():
        path_ = path + (k,)
        result.append(path_)

        if isinstance(v, dict):
            result.extend(tupler(v, path_))

        if isinstance(v, set):
            for v_ in v:
                result.append(path_ + (v_,))

    return result


def merge_trees(d1: dict, d2: dict) -> dict:
    """Recursively merge two tree-like dicts (values always dicts)."""
    result = dict(d1)  # shallow copy of d1
    for k, v in d2.items():
        if k in result:
            result[k] = merge_trees(
                result[k],
                v,
            )  # recurse since v must also be a dict
        else:
            result[k] = v
    return result


def dictify(d: defaultdict | dict) -> dict:
    """Recursively convert defaultdict to dict."""
    if isinstance(d, defaultdict):
        return {k: dictify(v) for k, v in d.items()}
    return d



"""Default Temporal Horizon"""

# from dataclasses import dataclass, field
# import calendar
# from typing import TypeAlias

# Rng: TypeAlias = tuple[tuple[int, int], tuple[int, int]] | tuple[tuple[int, int]]

# @dataclass
# class _Calendar:
#     rng_year: tuple[int, int] = field(default=None)
#     rng_month: Rng = field(
#         default=None
#     )
#     rng_week: Rng = field(
#         default=None
#     )
#     rng_day: Rng= field(
#         default=None
#     )
#     rng_hour: Rng = field(
#         default=None
#     )
#     rng_minute: Rng = field(
#         default=None
#     )
#     def_year: bool = field(default=False)
#     def_month: bool = field(default=False)
#     def_week: bool = field(default=False)
#     def_day: bool = field(default=False)
#     def_hour: bool = field(default=False)
#     def_minute: bool = field(default=False)

#     def __post_init__(self):
#         self._calendar = calendar.Calendar()
#         rng_attrs = [
#             'rng_year',
#             'rng_month',
#             'rng_week',
#             'rng_day',
#             'rng_hour',
#             'rng_minute',
#         ]

#         if sum(rng is not None for getattr(self, rng) in rng_attrs) > 1:
#             raise ValueError('Only one range attribute can be set')

#         rng = [getattr(self, rng) for rng in rng_attrs if getattr(self, rng) is not None]

#         for r in rng:
#             if self.range_year:
#                 year = r
#             else:
#                 year = r[0]

#             cal = self._calendar.yeardayscalendar(year, width=12)
#             if self.range_month:
#                 months = cal[r[1]]

#             if self.range_week:



"""Location where Operations can reside"""

from __future__ import annotations

from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING, Optional, Self
from warnings import warn

from gana import I as Idx

from ..._core._x import _X
from ...utils.dictionary import get_depth
from .linkage import Linkage

if TYPE_CHECKING:
    from ...dimensions.space import Space
    from ..commodity.currency import Currency
    from ..operation.process import Process
    from ..operation.storage import Storage


class Location(_X):
    """A discretization of Space
    A location can be inclusive of other locations


    :param has: Locations contained in this Location.
    :type has: tuple[IsLocation]
    :param label: Label of the Location. Defaults to None.
    :type label: str, optional

    :ivar model: Model to which the Location belongs.
    :vartype model: Model
    :ivar name: Name of the Location. Set when the Location is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the Location.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the Location.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the Location with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar isin: Location in which this Location is contained. Defaults to None.
    :vartype isin: IsLocation, optional
    :ivar currency: Currency used in the Location. Defaults to None.
    :vartype currency: Currency, optional
    :ivar alsohas: Locations contained in this Location, but not directly. Defaults to ()
    :vartype alsohas: tuple[IsLocation]
    :ivar hierarchy: Level of hierarchy of the Location. Defaults to None.
    :vartype hierarchy: int, optional
    """

    def __init__(self, *has: Self, label: str = None):

        # it is an indexed component
        _X.__init__(self, label=label or "")

        # the other locations contained in this location
        self.has: tuple[Self] = has
        # if the location is a part of another
        self.isin: Optional[Self] = None
        # the currency used in the location
        self.currency: Optional[Currency] = None
        # goes down another level of hierarchy
        # to find locations within the locations contained in this location
        self.alsohas: tuple[Self] = ()

        # Hierarchy in the space tree
        self.hierarchy: int = None

        for loc in self.has:
            if loc.name:
                # if the location has been set on the program
                # set the location as a part of this location
                loc.isin = self
                # set the location on to itself
                setattr(self, loc.name, loc)
                for locin in loc.has:
                    # if not already notified
                    # update alsohas
                    if locin not in self.has:
                        self.alsohas += (locin,)

    def __setattr__(self, name, value):

        if name == "currency" and value:
            # all locations within a location have the same currency
            for loc in self.has + self.alsohas:
                loc.currency = value

        super().__setattr__(name, value)

    def locate(self, *operations: Process | Storage):
        """Locates the Operations"""
        # locates an operation
        # which leads to the conversion balances being written
        for opr in operations:
            opr.locate(self)

    @property
    def tree(self) -> dict[Self, dict]:
        """Prints the tree of Locations"""
        # a tree visualizing the hierarchy of locations
        if self.has:
            return {loc: loc.tree for loc in self.has}
        # lowest level of the hierarchy locations
        # return an empty dictionary (set)
        return {}

    @property
    def depth(self):
        """Finds the depth of the Location"""
        # gets the level of hierarchy of the Location
        return get_depth(self.tree)

    def update_hierarchy(self, hierarchy: int = 0):
        """Updates the hierarchy of the locations"""

        self.hierarchy = hierarchy
        for loc in self.has:
            loc.update_hierarchy(hierarchy + 1)

    @property
    def space(self) -> Space:
        """Space to which the Location belongs"""
        return self.model.space

    @property
    def network(self) -> Self:
        """Network to which the Location belongs"""
        return self.model.network

    @property
    def isnetwork(self) -> bool:
        """Is this the network of the model?"""
        return self == self.model.network

    @cached_property
    def I(self) -> Idx:
        """gana index set"""
        # if self.has:
        #     # if component has a nested locations
        #     _index = Idx(*[loc.name for loc in self.has], tag=self.label or "")

        # else:
        # if component has no nested locations
        _index = Idx(self.name, tag=self.label or "")
        setattr(self.program, self.name, _index)
        return _index

    def sink(self):
        """Tells whether the location is a sink"""
        if self in self.space.sinks:
            return True
        return False

    def source(self):
        """Tells whether the location is a source"""
        if self in self.space.sources:
            return True
        return False

    def links(self, location, print_link: bool = True) -> list[Linkage]:
        """Finds the links between two Locations

        Args:
            location (IsLocation): Location to find links with
            print_link (bool, optional): Whether the links are to be printed. Defaults to True.

        Returns:
            list: Provides the links between the locations
        """
        # this prints out all the links between the two locations
        links = []
        for link in self.space.linkages:
            source, sink = False, False

            if is_(self, link.source) and is_(location, link.sink):
                source, sink = self, location

            if is_(self, link.sink) and is_(location, link.source):
                source, sink = location, self

            if source and sink:
                links.append(link)
                if print_link:
                    print(f"{source} is source and {sink} is sink in {link}")
                continue
        return links

    def connected(self, location, print_link: bool = False) -> bool:
        """Finds whether the Locations are connected

        Args:
            location (IsLocation): Location to verify Links with
            print_link (bool, optional): Whether to print the Links. Defaults to False.

        Returns:
            bool: True if Locations are connection
        """
        # this lets you know whether the two locations are connected
        if self.links(location, print_link=print_link):
            return True
        return False

    def all(self):
        """gives locations within"""
        # yeilds all locations within the location and all locations within those locations
        if self.has:
            for loc in self.has:
                yield loc
                yield from loc.all()

    def __add__(self, location: Self):
        """Creates another location which consists of self and other"""
        if not self.name:
            # this handles when adding multiple locations
            return Location(*self.has, location)
        return Location(location, self)

    def __radd__(self, number: int):
        """For allowing sum([Loc])"""
        if isinstance(number, int) and number == 0:
            return self

    def __sub__(self, location: Self):
        """Creates a linkage"""
        # subtracting a location creates a link
        # alternatively Model.Link can be used
        # for multiple links across the same two locations
        # declare Link() objects
        links = self.links(location, print_link=False)
        links = [link for link in links if link.source == self]
        if len(links) > 1:
            warn(
                f"Multiple links found between ({self}, {location})\n"
                "Suggest using model.named_link = Link(...)\n"
                f"Currently, taking {links[0]}, with distance {links[0].dist}",
                UserWarning,
            )
        if links:
            return links[0]

    def __eq__(self, other: Self):
        # the objects need to be the same
        return is_(self, other)

    def __lt__(self, other: Self):
        # if a location is nested within another
        return self in other.has or self in other.alsohas

    def __leq__(self, other: Self):
        # same as less than & equality
        return self < other or self == other

    def __gt__(self, other: Self):
        # reverse check
        return other < self

    def __geq__(self, other: Self):
        # reverse check
        return other <= self

    def __iter__(self):
        yield self.has

    def __contains__(self, item: Self):
        if item in self.has:
            return True



"""Space"""

from dataclasses import dataclass

from .._core._dimension import _Dimension
from ..components.spatial.linkage import Linkage
from ..components.spatial.location import Location


@dataclass
class Space(_Dimension):
    """
    Spatial representation of the system.

    All spatial components are attached to this object.


    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, auto generated
    :vartype name: str
    :ivar locations: List of locations in the space.
    :vartype locations: list[Loc]
    :ivar sources: List of source locations.
    :vartype sources: list[Loc]
    :ivar sinks: List of sink locations.
    :vartype sinks: list[Loc]
    :ivar linkages: List of linkages in the space.
    :vartype linkages: list[Link]
    :ivar label: Label for the space.
    :vartype label: str
    :ivar default: Default location for the space. Defaults to None.
    :vartype default: Loc
    :ivar network: The encompassing region (network) of the space.
    :vartype network: Loc
    :ivar s: List of spatial components (locations and linkages).
    :vartype s: list[Loc | Link]
    :ivar tree: Nested dictionary of locations.
    :vartype tree: dict
    :ivar hierarchy: position on tree.
    :vartype hierarchy: dict[int, list[Loc]]


    .. note::
        - name is self generated
        - locations, sources, sinks, and linkages are populated as model is defined
        - label is fixed
        - default is set to None initially and is updated when needed (see network property)
    """

    def __post_init__(self):
        self.locations: list[Location] = []
        self.sources: list[Location] = []
        self.sinks: list[Location] = []
        self.linkages: list[Linkage] = []

        _Dimension.__post_init__(self)

    # -----------------------------------------------------
    #                    Helpers
    # -----------------------------------------------------

    @property
    def tree(self) -> dict:
        """creates a nested dictionary of locations"""
        tree_ = {self.network: {}}

        for loc in self.network.has:
            tree_[self.network][loc] = loc.tree

        return tree_

    @property
    def hierarchy(self) -> dict[int, list[Location]]:
        """gives position in tree"""
        self.network.update_hierarchy()
        hierarchy_ = {}
        for spc in self.s:
            if spc.hierarchy not in hierarchy_:
                hierarchy_[spc.hierarchy] = []
            hierarchy_[spc.hierarchy].append(spc)
        return hierarchy_

    # -----------------------------------------------------
    #                    Superlative
    # -----------------------------------------------------

    @property
    def network(self) -> Location:
        """An encompassing location"""

        # if no location is available, create a default one
        if not self.locations:
            return self.model.default_location()

        # if only one location is available, return it
        if len(self.locations) == 1:
            return self.locations[0]

        # check for locations that are not nested
        loc_not_nested = [loc for loc in self.locations if not loc.isin]

        # only one implies that all locations are nested under the one location
        # which is the network
        if len(loc_not_nested) == 1:
            return loc_not_nested[0]

        # if multiple "not nested" locations exist, make a network with them
        if loc_not_nested:
            # l is used as the name for the default network
            self.model.ntw = sum(loc_not_nested)
            return self.model.ntw

        raise ValueError("No network location could be determined")

    @property
    def s(self) -> list[Location | Linkage]:
        """List of spatial components"""
        return self.locations + self.linkages

    def split(self, loc: Location) -> tuple[list[Location], list[Location]]:
        """Gives a list of locations at a higher and lower hierarchy than loc"""
        hierarchy = self.hierarchy

        loc_pos = loc.hierarchy

        if loc_pos + 1 in hierarchy:
            lower = [l for l in hierarchy[loc_pos + 1] if l in loc.has]
        else:
            lower = []

        if loc_pos - 1 in hierarchy:
            upper = [l for l in hierarchy[loc_pos - 1] if loc in l.has]

            upper = upper[0] if upper else None
        else:
            upper = None

        return lower, upper



"""Resource are:
1. converted by Processes
2. stored by Storage
3. transported by Transits
4. lost by Storage and Transits
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from energia.components.commodity._commodity import _Commodity
from energia.components.impact.categories import Environ

if TYPE_CHECKING:
    from ...modeling.constraints.calculate import Calculate


@dataclass
class Resource(_Commodity):
    """
    A resource, can be a material, chemical, energy, etc.

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """

    def __post_init__(self):
        _Commodity.__post_init__(self)

        # base resource, if any in conversion
        self.inv_of: Resource = None

        # resource in its stored form
        self.in_inv: list[Resource] = []

    # @property
    # def consume(self):
    #     return self.model.consume(self)

    @property
    def gwp(self) -> Calculate:
        """Global Warming Potential"""
        if not hasattr(self.model, "GWP"):
            self.model.GWP = Environ(label="Global Warming Potential (kg CO2)")

        return self.consume[self.model.GWP.emit]

    @property
    def htp(self) -> Calculate:
        """Human Toxicity Potential"""
        if not hasattr(self.model, "HTP"):
            self.model.HTP = Environ(label="Human Toxicity Potential (kg 1,4-DB eq.)")

        return self.consume[self.model.HTP.emit]

    @property
    def demand(self) -> Calculate:
        """Demand (alias for the Aspect release)"""
        return self.release

    @property
    def price(self) -> Calculate:
        """Cost of consume"""
        return self.consume[self.model.default_currency().spend]



"""Linkage links Locations through Transits"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING, Self

from gana import I

from ..._core._x import _X

if TYPE_CHECKING:
    from ...dimensions.space import Space
    from ..measure.unit import Unit
    from .location import Location


@dataclass
class Linkage(_X):
    """
    Linkage between two Locations.

    :param name: Name of the Linkage. Defaults to None.
    :type name: str, optional
    :param label: Label of the Linkage. Defaults to None.
    :type label: str, optional
    :param source: Source Location.
    :type source: Loc
    :param sink: Sink Location.
    :type sink: Loc
    :param dist: Distance between the two Locations.
    :type dist: float | Unit
    :param bi: Is the Linkage bidirectional? Defaults to False.
    :type bi: bool, optional
    :param auto: Is the Linkage automatically generated? Defaults to False.
    :type auto: bool, optional

    :ivar model: Model to which the Linkage belongs.
    :vartype model: Model
    :ivar space: Space to which the Linkage belongs.
    :vartype space: Space
    :ivar network: Network to which the Linkage belongs.
    :vartype network: Loc
    """

    source: Location = None
    sink: Location = None
    dist: float | Unit = None
    basis: Unit = None
    bi: bool = False
    auto: bool = False

    def __post_init__(self):

        if is_(self.source, self.sink):
            # if the source and sink are the same, throw error
            raise ValueError(f"source and sink can't both be {self.source}")

        _X.__post_init__(self)
        self.sib: Self = None

        self.hierarchy = 1

    @cached_property
    def space(self) -> Space:
        """Space to which the Location belongs"""
        return self.model.space

    @property
    def network(self) -> Location:
        """Network to which the Location belongs"""
        return self.model.network

    @property
    def isnetwork(self) -> bool:
        """Is this the network of the model?
        Linkage can never be the network
        """
        return False

    @property
    def isin(self) -> Location | None:
        """Location to which the Linkage belongs"""
        if self.source.isin:
            if self.source.isin == self.sink.isin:
                self.source.isin.has += (self,)
                return self.source.isin
        self.network.has += (self,)
        return self.network

    def rev(self):
        """Reversed Link"""
        # locations are all uni directional
        # if bi is True, create another link for the other direction
        # and set self.bi to False
        if self.bi:
            self.bi = False
            return -self

    def __neg__(self):
        # creates a new Link in the opposite direction
        if self.sib is not None:
            return self.sib

        if self.label:
            label = self.label + "(-)"
            self.label = self.label + "(+)"
        else:
            label = None

        _link = Linkage(
            source=self.sink,
            sink=self.source,
            dist=self.dist,
            label=label,
        )
        if self.auto:
            _link.name = f"{self.sink.name}-{self.source.name}"
        else:
            _link.name = "-" + self.name
        _link.sib, self.sib = self, _link
        return _link

    def __eq__(self, other: int | float | Self) -> Self | bool:
        """Sets the distance of a linkage"""
        # sets the distance of a linkage
        # each linkage has a unique distance
        if isinstance(other, (int, float)):

            self.dist = other

            self.model = self.source.model

            if not self.name:
                self.name = rf"{self.source.name}_{self.sink.name}"

            setattr(self.model, self.name, self)

            return self

        if is_(self, other):
            return True
        return False

    @cached_property
    def I(self) -> I:
        """gana index set (I)"""
        _index = I(self.name, tag=self.label or "")
        setattr(self.program, self.name, _index)
        return _index

    def update_hierarchy(self, hierarchy: int = 0):
        """Updates the hierarchy of the locations"""

        self.hierarchy = hierarchy
